"""
The DUL's finite state machine representation.
"""
import logging
import queue
from typing import TYPE_CHECKING, cast

from pynetdicom import evt
from pynetdicom.pdu import (
    A_ASSOCIATE_RQ,
    A_ASSOCIATE_RJ,
    A_ASSOCIATE_AC,
    P_DATA_TF,
    A_RELEASE_RQ,
    A_RELEASE_RP,
    A_ABORT_RQ,
)
from pynetdicom.pdu_primitives import A_P_ABORT, A_ABORT
from pynetdicom.transport import T_CONNECT

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.dul import DULServiceProvider
    from pynetdicom.transport import AssociationSocket
    from pynetdicom.pdu_primitives import A_ASSOCIATE, P_DATA, A_RELEASE


LOGGER = logging.getLogger("pynetdicom.sm")


class InvalidEventError(Exception):
    """Exception for use when an invalid event occurs for a given state."""

    pass


# pylint: disable=invalid-name
class StateMachine:
    """Implementation of the DICOM Upper Layer State Machine.

    Attributes
    ----------
    current_state : str
        The current state of the state machine, ``'Sta1'`` to ``'Sta13'``.
    dul : dul.DULServiceProvider
        The DICOM Upper Layer service instance for the local AE

    References
    ----------

    * DICOM Standard, Part 8, :dcm:`Section 9.2<part08/sect_9.2.html>`
    """

    def __init__(self, dul: "DULServiceProvider"):
        """Create a new :class:`StateMachine`.

        Parameters
        ---------
        dul : dul.DULServiceProvider
            The DICOM Upper Layer Service instance for the association.
        """
        self.current_state = "Sta1"
        self.dul = dul

    def do_action(self, event: str) -> None:
        """Execute the action triggered by `event`.

        Parameters
        ----------
        event : str
            The event to be processed, ``'Evt1'`` to ``'Evt19'``
        """
        # Check (event + state) is valid
        if (event, self.current_state) not in TRANSITION_TABLE:
            msg = "Invalid event '{}' for the current state '{}'".format(
                event, self.current_state
            )
            LOGGER.error(msg)
            raise InvalidEventError(msg)

        action_name = TRANSITION_TABLE[(event, self.current_state)]

        # action is the (description, function, state) tuple
        #   associated with the action_name
        action = ACTIONS[action_name]

        # Attempt to execute the action and move the state machine to its
        #   next state
        try:
            # Execute the required action
            next_state = action[1](self.dul)

            # Event handler - FSM transition
            evt.trigger(
                self.dul.assoc,
                evt.EVT_FSM_TRANSITION,
                {
                    "action": action_name,
                    "current_state": self.current_state,
                    "fsm_event": event,
                    "next_state": next_state,
                },
            )
            # print(
            #     "{}: {} + {} -> {} -> {}".format(
            #         self.dul.assoc.mode[0].upper(),
            #         self.current_state,
            #         event,
            #         action_name,
            #         next_state,
            #     )
            # )

            # Move the state machine to the next state
            self.transition(next_state)

        except Exception as exc:
            LOGGER.error(
                "State Machine received an exception attempting "
                "to perform the action '%s' while in state '%s'",
                action_name,
                self.current_state,
            )
            LOGGER.exception(exc)
            self.dul.kill_dul()
            raise

    def transition(self, state: str) -> None:
        """Transition the state machine to the next state.

        Parameters
        ----------
        state : str
            The state to transition to, ``'Sta1'`` to ``'Sta13'``.

        Raises
        ------
        ValueError
            If `state` is not a valid state.
        """
        # Validate that state is acceptable
        if state in STATES.keys():
            self.current_state = state
        else:
            msg = f"Invalid state '{state}' for State Machine"
            LOGGER.error(msg)
            raise ValueError(msg)


def AE_1(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-1.

    *Event*

    Service user issued A-ASSOCIATE (request) to the service provider

    *Action*

    Issue TRANSPORT CONNECT request primitive to the transport service.

    *State/Event Triggers*

    - Sta1 + Evt1

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta4'``, the next state of the state machine.
    """
    # A-ASSOCIATE (request) primitive received from local user
    request = cast("A_ASSOCIATE", dul.to_provider_queue.get(False))

    # Issue TRANSPORT CONNECT request primitive to local transport service
    sock = cast("AssociationSocket", dul.socket)
    sock.connect(T_CONNECT(request))

    return "Sta4"


def AE_2(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-2.

    On receiving connection confirmation, send A-ASSOCIATE-RQ to the peer AE
    This send a byte stream with the format given by Table 9-11

    State-event triggers: Sta4 + Evt2

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta5'``, the next state of the state machine.
    """
    # TRANSPORT CONNECTION primitive received from transport service
    primitive = cast("T_CONNECT", dul.to_provider_queue.get(False))

    # Send A-ASSOCIATE-RQ PDU to the peer
    dul._send(A_ASSOCIATE_RQ(primitive.request))

    return "Sta5"


def AE_3(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-3.

    On receiving A-ASSOCIATE-AC, issue acceptance confirmation

    State-event triggers: Sta5 + Evt3

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta6'``, the next state of the state machine
    """
    # Received A-ASSOCIATE-AC PDU from the peer
    pdu = dul._recv_pdu.get(False)

    # Issue A-ASSOCIATE confirmation (accept) primitive
    dul.to_user_queue.put(cast("A_ASSOCIATE", pdu.to_primitive()))

    return "Sta6"


def AE_4(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-4.

    On receiving A-ASSOCIATE-RJ, issue rejection confirmation and close
    connection

    State-event triggers: Sta5 + Evt4

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    # Received A-ASSOCIATE-RJ PDU from the peer
    pdu = dul._recv_pdu.get(False)

    # Issue A-ASSOCIATE confirmation (reject) primitive and close transport
    # connection
    dul.to_user_queue.put(cast("A_ASSOCIATE", pdu.to_primitive()))
    sock = cast("AssociationSocket", dul.socket)
    sock.close()

    assoc = dul.assoc
    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor

    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    dul.kill_dul()

    return "Sta1"


def AE_5(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-5.

    From Idle state, on receiving a remote connection attempt, respond and
    start ARTIM. This is the first step in associating a remote AE (requestor)
    to the local AE (acceptor).

    State-event triggers: Sta1 + Evt5

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta2'``, the next state of the state machine
    """
    # Start ARTIM timer
    dul.artim_timer.start()

    return "Sta2"


def AE_6(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-6.

    On receiving an A-ASSOCIATE-RQ PDU from the peer then stop the ARTIM timer
    and then either

    * issue an A-ASSOCIATE indication primitive if the -RQ is acceptable or
    * issue an A-ASSOCIATE-RJ PDU to the peer and start the ARTIM timer

    This is a lower-level DUL Service Provider initiated rejection - for
    example this could be where the protocol version is checked

    State-event triggers: Sta2 + Evt6

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        Either ``'Sta3'`` or ``'Sta13'``, the next state of the state machine
    """
    # Stop ARTIM timer
    dul.artim_timer.stop()

    recv_pdu = cast(A_ASSOCIATE_RQ, dul._recv_pdu.get(False))
    primitive = recv_pdu.to_primitive()

    # If A-ASSOCIATE-RQ not acceptable by service dul provider
    #   Then set reason and send -RJ PDU back to peer
    if recv_pdu.protocol_version != 0x0001:
        LOGGER.error(
            "A-ASSOCIATE-RQ: Unsupported protocol version "
            f"'0x{recv_pdu.protocol_version:04X}'"
        )

        # Send A-ASSOCIATE-RJ PDU and start ARTIM timer
        primitive.result = 0x01
        primitive.result_source = 0x02
        primitive.diagnostic = 0x02

        dul._send(A_ASSOCIATE_RJ(primitive))

        dul.artim_timer.start()

        return "Sta13"

    # If A-ASSOCIATE-RQ acceptable by service dul provider
    #   issue A-ASSOCIATE indication primitive and move to Sta3
    dul.to_user_queue.put(primitive)

    return "Sta3"


def AE_7(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-7.

    On receiving association request acceptance, issue A-ASSOCIATE-AC

    State-event triggers: Sta3 + Evt7

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta6'``, the next state of the state machine
    """
    # Received A-ASSOCIATE (AC) primitive from local user
    primitive = cast("A_ASSOCIATE", dul.to_provider_queue.get(False))

    # Send A-ASSOCIATE-AC PDU
    dul._send(A_ASSOCIATE_AC(primitive))

    return "Sta6"


def AE_8(dul: "DULServiceProvider") -> str:
    """Association establishment action AE-8.

    On receiving association request rejection, issue A-ASSOCIATE-RJ

    State-event triggers: Sta3 + Evt8

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    # Received A-ASSOCIATE (RJ) primitive from local user
    primitive = cast("A_ASSOCIATE", dul.to_provider_queue.get(False))

    # Send A-ASSOCIATE-RJ PDU and start ARTIM timer
    dul._send(A_ASSOCIATE_RJ(primitive))
    dul.artim_timer.start()

    return "Sta13"


def DT_1(dul: "DULServiceProvider") -> str:
    """Data transfer DT-1.

    On receiving a P-DATA request, send P-DATA-TF

    State-event triggers: Sta6 + Evt9

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta6'``, the next state of the state machine
    """
    # P-DATA request received from local user
    primitive = cast("P_DATA", dul.to_provider_queue.get(False))

    # Send P-DATA-TF PDU
    dul._send(P_DATA_TF(primitive))

    return "Sta6"


def DT_2(dul: "DULServiceProvider") -> str:
    """Data transfer DT-2.

    On receiving a P-DATA-TF request, send P-DATA indication

    State-event triggers: Sta6 + Evt10

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta6'``, the next state of the state machine
    """
    # P-DATA-TF PDU received from peer
    pdu = dul._recv_pdu.get(False)

    # Send P-DATA indication primitive directly to DIMSE for processing
    dul.assoc.dimse.receive_primitive(cast("P_DATA", pdu.to_primitive()))

    return "Sta6"


def AR_1(dul: "DULServiceProvider") -> str:
    """Association release AR-1.

    Send Association release request

    State-event triggers: Sta6 + Evt11

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta7'``, the next state of the state machine
    """
    # Received A-RELEASE request from local user
    primitive = cast("A_RELEASE", dul.to_provider_queue.get(False))

    # Send A-RELEASE-RQ PDU
    dul._send(A_RELEASE_RQ(primitive))

    return "Sta7"


def AR_2(dul: "DULServiceProvider") -> str:
    """Association release AR-2.

    On receiving an association release request, send release indication

    State-event triggers: Sta6 + Evt12

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta8'``, the next state of the state machine
    """
    # A-RELEASE-RQ PDU received from peer
    pdu = dul._recv_pdu.get(False)

    # Send A-RELEASE indication primitive
    dul.to_user_queue.put(pdu.to_primitive())

    return "Sta8"


def AR_3(dul: "DULServiceProvider") -> str:
    """Association release AR-3.

    On receiving an association release response, send release confirmation,
    close connection and go back to Idle state

    State-event triggers: Sta7 + Evt13, Sta11 + Evt13

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    # A-RELEASE-RP PDU received from peer
    pdu = dul._recv_pdu.get(False)

    # Issue A-RELEASE confirmation primitive and close transport connection
    dul.to_user_queue.put(pdu.to_primitive())
    sock = cast("AssociationSocket", dul.socket)
    sock.close()

    assoc = dul.assoc
    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor

    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    dul.kill_dul()

    return "Sta1"


def AR_4(dul: "DULServiceProvider") -> str:
    """Association release AR-4.

    On receiving an association release response, send release response

    State-event triggers: Sta8 + Evt14, Sta12 + Evt14

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    # A-RELEASE (response) primitive received from local user
    primitive = cast("A_RELEASE", dul.to_provider_queue.get(False))

    # Issue A-RELEASE-RP PDU and start ARTIM timer
    dul._send(A_RELEASE_RP(primitive))
    dul.artim_timer.start()

    return "Sta13"


def AR_5(dul: "DULServiceProvider") -> str:
    """Association release AR-5.

    On receiving transport connection closed, stop the ARTIM timer and go back
    to Idle state

    State-event triggers: Sta13 + Evt17

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    assoc = dul.assoc
    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor

    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    # Stop ARTIM timer
    dul.artim_timer.stop()
    dul.kill_dul()

    return "Sta1"


def AR_6(dul: "DULServiceProvider") -> str:
    """Association release AR-6.

    On receiving P-DATA-TF during attempted association release request
    send P-DATA indication

    State-event triggers: Sta7 + Evt10

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta7'``, the next state of the state machine
    """
    # P-DATA-TF PDU received from peer
    pdu = cast("P_DATA_TF", dul._recv_pdu.get(False))

    # Issue P-DATA indication
    dul.to_user_queue.put(pdu.to_primitive())

    return "Sta7"


def AR_7(dul: "DULServiceProvider") -> str:
    """Association release AR-7.

    On receiving P-DATA request during attempted association release request
    send P-DATA-TF

    State-event triggers: Sta8 + Evt9

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta8'``, the next state of the state machine
    """
    # P-DATA primitive received from local user
    primitive = cast("P_DATA", dul.to_provider_queue.get(False))

    # Issue P-DATA-TF PDU
    dul._send(P_DATA_TF(primitive))

    return "Sta8"


def AR_8(dul: "DULServiceProvider") -> str:
    """Association release AR-8.

    On receiving association release request while local is requesting release
    then issue release collision indication

    State-event triggers: Sta7 + Evt12

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        Either ``'Sta9'`` or ``'Sta10'``, the next state of the state machine
    """
    # A-RELEASE-RQ PDU received from peer
    pdu = cast("A_RELEASE_RQ", dul._recv_pdu.get(False))

    # Issue A-RELEASE indication (release collision)
    dul.to_user_queue.put(pdu.to_primitive())
    if dul.assoc.is_requestor:
        return "Sta9"

    return "Sta10"


def AR_9(dul: "DULServiceProvider") -> str:
    """Association release AR-9.

    On receiving A-RELEASE primitive, send release response

    State-event triggers: Sta9 + Evt14

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta11'``, the next state of the state machine
    """
    # A-RELEASE (response) primitive received from local user
    primitive = cast("A_RELEASE", dul.to_provider_queue.get(False))

    # Send A-RELEASE-RP PDU
    dul._send(A_RELEASE_RP(primitive))

    return "Sta11"


def AR_10(dul: "DULServiceProvider") -> str:
    """Association release AR-10.

    On receiving A-RELEASE-RP, issue release confirmation

    State-event triggers: Sta10 + Evt13

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta12'``, the next state of the state machine
    """
    # A-RELEASE-RP PDU received from peer
    pdu = cast("A_RELEASE_RP", dul._recv_pdu.get(False))

    # Issue A-RELEASE confirmation primitive
    dul.to_user_queue.put(pdu.to_primitive())

    return "Sta12"


def AA_1(dul: "DULServiceProvider") -> str:
    """Association abort AA-1.

    If on sending A-ASSOCIATE-RQ we receive an invalid reply, or an abort
    request then abort

    State-event triggers: Sta2 + Evt3/Evt4/Evt10/Evt12/Evt13/Evt19,
    Sta3/Sta5/Sta6/Sta7/Sta8/Sta9/Sta10/Sta11/Sta12 + Evt15

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    # Received invalid PDU from peer or an A-ABORT primitive from local user
    try:
        primitive = dul.to_provider_queue.queue[0]
        if isinstance(primitive, (A_ABORT, A_P_ABORT)):
            primitive = dul.to_provider_queue.get(False)
    except (queue.Empty, IndexError):
        primitive = None

    # Send A-ABORT PDU (service-user source) and start (or restart
    # if already started) ARTIM timer.
    pdu = A_ABORT_RQ()
    if primitive is not None:
        pdu.from_primitive(primitive)
    else:
        # Reason not specified
        pdu.source = 0x00
        pdu.reason_diagnostic = 0x00

    dul._send(pdu)
    dul.artim_timer.restart()

    return "Sta13"


def AA_2(dul: "DULServiceProvider") -> str:
    """Association abort AA-2.

    On receiving an A-ABORT or if the ARTIM timer expires, close connection and
    return to Idle

    State-event triggers: Sta2 + Evt16/Evt18, Sta4 + Evt15, Sta13 + Evt16/Evt18

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    # Stop ARTIM timer if running. Close transport connection.
    dul.artim_timer.stop()
    sock = cast("AssociationSocket", dul.socket)
    sock.close()

    assoc = dul.assoc
    assoc.dimse.msg_queue.put((None, None))

    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor
    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    dul.kill_dul()

    return "Sta1"


def AA_3(dul: "DULServiceProvider") -> str:
    """Association abort AA-3.

    On receiving A-ABORT, issue abort indication, close connection and
    return to Idle

    State-event triggers: Sta3/Sta5/Sta6/Sta7/Sta8/Sta9/Sta10/Sta11/Sta12 +
    Evt16

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    # A-ABORT PDU received from peer
    pdu = cast("A_ABORT_RQ", dul._recv_pdu.get(False))

    # If (service-user initiated abort):
    #   - Issue A-ABORT indication and close transport connection.
    # Otherwise (service-dul initiated abort):
    #   - Issue A-P-ABORT indication and close transport connection.
    # This action is triggered by the reception of an A-ABORT PDU
    dul.to_user_queue.put(pdu.to_primitive())
    sock = cast("AssociationSocket", dul.socket)
    sock.close()

    assoc = dul.assoc
    assoc.dimse.msg_queue.put((None, None))

    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor
    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    dul.kill_dul()

    return "Sta1"


def AA_4(dul: "DULServiceProvider") -> str:
    """Association abort AA-4.

    If connection closed, issue A-P-ABORT and return to Idle

    State-event triggers: Sta3/Sta4/Sta5/Sta6/Sta7/Sta8/Sta9/Sta10/Sta11/Sta12
    + Evt17

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    assoc = dul.assoc
    assoc.dimse.msg_queue.put((None, None))

    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor
    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    # Issue A-P-ABORT indication primitive.
    primitive = A_P_ABORT()
    primitive.provider_reason = 0x00
    dul.to_user_queue.put(primitive)
    dul.kill_dul()

    return "Sta1"


def AA_5(dul: "DULServiceProvider") -> str:
    """Association abort AA-5.

    If connection closed during association request, stop ARTIM timer and
    return to Idle

    State-event triggers: Sta2 + Evt17

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta1'``, the next state of the state machine
    """
    assoc = dul.assoc
    remote = assoc.acceptor if assoc.is_requestor else assoc.requestor

    address = (remote.address, remote.port)
    evt.trigger(dul.assoc, evt.EVT_CONN_CLOSE, {"address": address})

    # Stop ARTIM timer.
    dul.artim_timer.stop()
    dul.kill_dul()

    return "Sta1"


def AA_6(dul: "DULServiceProvider") -> str:
    """Association abort AA-6.

    If receive a PDU while waiting for connection to close, ignore it

    State-event triggers: Sta13 + Evt3/Evt4/Evt10/Evt12/Evt13

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    # Ignore PDU
    try:
        dul._recv_pdu.get(False)
    except queue.Empty:
        pass

    return "Sta13"


def AA_7(dul: "DULServiceProvider") -> str:
    """Association abort AA-7.

    If receive a association request or invalid PDU while waiting for
    connection to close, send A-ABORT PDU

    State-event triggers: Sta13 + Evt6/Evt19

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    primitive = A_P_ABORT()
    primitive.provider_reason = 0x02

    # Send A-ABORT PDU
    dul._send(A_ABORT_RQ(primitive))

    return "Sta13"


def AA_8(dul: "DULServiceProvider") -> str:
    """Association abort AA-8.

    If receive invalid event, send A-ABORT PDU, issue A-P-ABORT indication and
    start ARTIM timer

    State-event triggers: Evt3 + Sta3/6/7/8/9/10/11/12,
    Evt4 + Sta3/5/6/7/8/9/10/11/12, Evt6 + Sta3/5/6/7/8/9/10/11/12,
    Evt10 + Sta3/5/8/9/10/11/12, Evt12 + Sta3/5/8/9/10/11/12,
    Evt13 + Sta3/5/6/8/9/12, Evt19 + Sta3/5/6/7/8/9/10/11/12

    Parameters
    ----------
    dul : dul.DULServiceProvider
        The DICOM Upper Layer Service instance for the local AE

    Returns
    -------
    str
        ``'Sta13'``, the next state of the state machine
    """
    # Send A-ABORT PDU (service-dul source), issue A-P-ABORT
    # indication, and start ARTIM timer.
    pdu = A_ABORT_RQ()
    pdu.source = 0x02  # A-P-ABORT
    pdu.reason_diagnostic = 0x00

    # Send A-ABORT PDU
    dul._send(pdu)

    # Issue A-P-ABORT to user
    primitive = A_P_ABORT()
    primitive.provider_reason = 0x05
    dul.to_user_queue.put(primitive)
    dul.artim_timer.start()

    return "Sta13"


# Finite State Machine
# Machine State Definitions, PS3.8 Tables 9-1, 9-2, 9-3, 9-4, 9-5
# pylint: disable=line-too-long
STATES = {
    # No association
    "Sta1": "Idle",
    # Association establishment
    "Sta2": "Transport connection open (Awaiting A-ASSOCIATE-RQ PDU)",
    "Sta3": "Awaiting local A-ASSOCIATE response primitive (from local user)",
    "Sta4": "Awaiting transport connection opening to complete (from local transport service)",
    "Sta5": "Awaiting A-ASSOCIATE-AC or A-ASSOCIATE-RJ PDU",
    # Data transfer
    "Sta6": "Association established and ready for data transfer",
    # Association release
    "Sta7": "Awaiting A-RELEASE-RP PDU",
    "Sta8": "Awaiting local A-RELEASE response primitive (from local user)",
    "Sta9": "Release collision requestor side; awaiting A-RELEASE response (from local user)",
    "Sta10": "Release collision acceptor side; awaiting A-RELEASE-RP PDU",
    "Sta11": "Release collision requestor side; awaiting A-RELEASE-RP PDU",
    "Sta12": "Release collision acceptor side; awaiting A-RELEASE response primitive (from local user)",
    # Waiting for transport connection close
    "Sta13": "Awaiting Transport Connection Close Indication (Association no longer exists)",
}

# State Machine Action Definitions, PS3.8 Tables 9-6, 9-7, 9-8, 9-9
ACTIONS = {
    # Association establishment related actions
    "AE-1": (
        "Issue TRANSPORT CONNECT request primitive to local transport service",
        AE_1,
        "Sta4",
    ),
    "AE-2": ("Send A-ASSOCIATE-RQ-PDU", AE_2, "Sta5"),
    "AE-3": ("Issue A-ASSOCIATE confirmation (accept) primitive", AE_3, "Sta6"),
    "AE-4": (
        "Issue A-ASSOCIATE confirmation (reject) primitive and close "
        "transport connection",
        AE_4,
        "Sta1",
    ),
    "AE-5": (
        "Issue Transport connection response primitive; start ARTIM timer",
        AE_5,
        "Sta2",
    ),
    "AE-6": (
        "Stop ARTIM timer and if A-ASSOCIATE-RQ acceptable by "
        "service-dul: issue A-ASSOCIATE indication primitive "
        "otherwise issue A-ASSOCIATE-RJ-PDU and start ARTIM timer",
        AE_6,
        ("Sta3", "Sta13"),
    ),
    "AE-7": ("Send A-ASSOCIATE-AC PDU", AE_7, "Sta6"),
    "AE-8": ("Send A-ASSOCIATE-RJ PDU and start ARTIM timer", AE_8, "Sta13"),
    # Data transfer related actions
    "DT-1": ("Send P-DATA-TF PDU", DT_1, "Sta6"),
    "DT-2": ("Send P-DATA indication primitive", DT_2, "Sta6"),
    # Association Release related actions
    "AR-1": ("Send A-RELEASE-RQ PDU", AR_1, "Sta7"),
    "AR-2": ("Issue A-RELEASE indication primitive", AR_2, "Sta8"),
    "AR-3": (
        "Issue A-RELEASE confirmation primitive and close transport connection",
        AR_3,
        "Sta1",
    ),
    "AR-4": ("Issue A-RELEASE-RP PDU and start ARTIM timer", AR_4, "Sta13"),
    "AR-5": ("Stop ARTIM timer", AR_5, "Sta1"),
    "AR-6": ("Issue P-DATA indication", AR_6, "Sta7"),
    "AR-7": ("Issue P-DATA-TF PDU", AR_7, "Sta8"),
    "AR-8": (
        "Issue A-RELEASE indication (release collision): if "
        "association-requestor, next state is Sta9, if not next "
        "state is Sta10",
        AR_8,
        ("Sta9", "Sta10"),
    ),
    "AR-9": ("Send A-RELEASE-RP PDU", AR_9, "Sta11"),
    "AR-10": ("Issue A-RELEASE confimation primitive", AR_10, "Sta12"),
    # Association abort related actions
    "AA-1": (
        "Send A-ABORT PDU (service-user source) and start (or "
        "restart if already started) ARTIM timer",
        AA_1,
        "Sta13",
    ),
    "AA-2": ("Stop ARTIM timer if running. Close transport connection", AA_2, "Sta1"),
    "AA-3": (
        "If (service-user initiated abort): issue A-ABORT "
        "indication and close transport connection, otherwise "
        "(service-dul initiated abort): issue A-P-ABORT indication "
        "and close transport connection",
        AA_3,
        "Sta1",
    ),
    "AA-4": ("Issue A-P-ABORT indication primitive", AA_4, "Sta1"),
    "AA-5": ("Stop ARTIM timer", AA_5, "Sta1"),
    "AA-6": ("Ignore PDU", AA_6, "Sta13"),
    "AA-7": ("Send A-ABORT PDU", AA_7, "Sta13"),
    "AA-8": (
        "Send A-ABORT PDU (service-dul source), issue an A-P-ABORT "
        "indication and start ARTIM timer",
        AA_8,
        "Sta13",
    ),
}

# State Machine Event Definitions, PS3.8 Table 9-10
EVENTS = {
    "Evt1": "A-ASSOCIATE request (local user)",
    "Evt2": "Transport connect confirmation (local transport service)",
    "Evt3": "A-ASSOCIATE-AC PDU (received on transport connection)",
    "Evt4": "A-ASSOCIATE-RJ PDU (received on transport connection)",
    "Evt5": "Transport connection indication (local transport service)",
    "Evt6": "A-ASSOCIATE-RQ PDU (on transport connection)",
    "Evt7": "A-ASSOCIATE response primitive (accept)",
    "Evt8": "A-ASSOCIATE response primitive (reject)",
    "Evt9": "P-DATA request primitive",
    "Evt10": "P-DATA-TF PDU (on transport connection)",
    "Evt11": "A-RELEASE request primitive",
    "Evt12": "A-RELEASE-RQ PDU (on transport)",
    "Evt13": "A-RELEASE-RP PDU (on transport)",
    "Evt14": "A-RELEASE response primitive",
    "Evt15": "A-ABORT request primitive",
    "Evt16": "A-ABORT PDU (on transport)",
    "Evt17": "Transport connection closed indication (local transport service)",
    "Evt18": "ARTIM timer expired (Association reject/release timer)",
    "Evt19": "Unrecognized or invalid PDU received",
}

# State Machine Transitions, PS3.8 Table 9-10
TRANSITION_TABLE = {
    ("Evt1", "Sta1"): "AE-1",
    ("Evt2", "Sta4"): "AE-2",
    ("Evt3", "Sta2"): "AA-1",
    ("Evt3", "Sta3"): "AA-8",
    ("Evt3", "Sta5"): "AE-3",
    ("Evt3", "Sta6"): "AA-8",
    ("Evt3", "Sta7"): "AA-8",
    ("Evt3", "Sta8"): "AA-8",
    ("Evt3", "Sta9"): "AA-8",
    ("Evt3", "Sta10"): "AA-8",
    ("Evt3", "Sta11"): "AA-8",
    ("Evt3", "Sta12"): "AA-8",
    ("Evt3", "Sta13"): "AA-6",
    ("Evt4", "Sta2"): "AA-1",
    ("Evt4", "Sta3"): "AA-8",
    ("Evt4", "Sta5"): "AE-4",
    ("Evt4", "Sta6"): "AA-8",
    ("Evt4", "Sta7"): "AA-8",
    ("Evt4", "Sta8"): "AA-8",
    ("Evt4", "Sta9"): "AA-8",
    ("Evt4", "Sta10"): "AA-8",
    ("Evt4", "Sta11"): "AA-8",
    ("Evt4", "Sta12"): "AA-8",
    ("Evt4", "Sta13"): "AA-6",
    ("Evt5", "Sta1"): "AE-5",
    ("Evt6", "Sta2"): "AE-6",
    ("Evt6", "Sta3"): "AA-8",
    ("Evt6", "Sta5"): "AA-8",
    ("Evt6", "Sta6"): "AA-8",
    ("Evt6", "Sta7"): "AA-8",
    ("Evt6", "Sta8"): "AA-8",
    ("Evt6", "Sta9"): "AA-8",
    ("Evt6", "Sta10"): "AA-8",
    ("Evt6", "Sta11"): "AA-8",
    ("Evt6", "Sta12"): "AA-8",
    ("Evt6", "Sta13"): "AA-7",
    ("Evt7", "Sta3"): "AE-7",
    ("Evt8", "Sta3"): "AE-8",
    ("Evt9", "Sta6"): "DT-1",
    ("Evt9", "Sta8"): "AR-7",
    ("Evt10", "Sta2"): "AA-1",
    ("Evt10", "Sta3"): "AA-8",
    ("Evt10", "Sta5"): "AA-8",
    ("Evt10", "Sta6"): "DT-2",
    ("Evt10", "Sta7"): "AR-6",
    ("Evt10", "Sta8"): "AA-8",
    ("Evt10", "Sta9"): "AA-8",
    ("Evt10", "Sta10"): "AA-8",
    ("Evt10", "Sta11"): "AA-8",
    ("Evt10", "Sta12"): "AA-8",
    ("Evt10", "Sta13"): "AA-6",
    ("Evt11", "Sta6"): "AR-1",
    ("Evt12", "Sta2"): "AA-1",
    ("Evt12", "Sta3"): "AA-8",
    ("Evt12", "Sta5"): "AA-8",
    ("Evt12", "Sta6"): "AR-2",
    ("Evt12", "Sta7"): "AR-8",
    ("Evt12", "Sta8"): "AA-8",
    ("Evt12", "Sta9"): "AA-8",
    ("Evt12", "Sta10"): "AA-8",
    ("Evt12", "Sta11"): "AA-8",
    ("Evt12", "Sta12"): "AA-8",
    ("Evt12", "Sta13"): "AA-6",
    ("Evt13", "Sta2"): "AA-1",
    ("Evt13", "Sta3"): "AA-8",
    ("Evt13", "Sta5"): "AA-8",
    ("Evt13", "Sta6"): "AA-8",
    ("Evt13", "Sta7"): "AR-3",
    ("Evt13", "Sta8"): "AA-8",
    ("Evt13", "Sta9"): "AA-8",
    ("Evt13", "Sta10"): "AR-10",
    ("Evt13", "Sta11"): "AR-3",
    ("Evt13", "Sta12"): "AA-8",
    ("Evt13", "Sta13"): "AA-6",
    ("Evt14", "Sta8"): "AR-4",
    ("Evt14", "Sta9"): "AR-9",
    ("Evt14", "Sta12"): "AR-4",
    ("Evt15", "Sta3"): "AA-1",
    ("Evt15", "Sta4"): "AA-2",
    ("Evt15", "Sta5"): "AA-1",
    ("Evt15", "Sta6"): "AA-1",
    ("Evt15", "Sta7"): "AA-1",
    ("Evt15", "Sta8"): "AA-1",
    ("Evt15", "Sta9"): "AA-1",
    ("Evt15", "Sta10"): "AA-1",
    ("Evt15", "Sta11"): "AA-1",
    ("Evt15", "Sta12"): "AA-1",
    ("Evt16", "Sta2"): "AA-2",
    ("Evt16", "Sta3"): "AA-3",
    ("Evt16", "Sta5"): "AA-3",
    ("Evt16", "Sta6"): "AA-3",
    ("Evt16", "Sta7"): "AA-3",
    ("Evt16", "Sta8"): "AA-3",
    ("Evt16", "Sta9"): "AA-3",
    ("Evt16", "Sta10"): "AA-3",
    ("Evt16", "Sta11"): "AA-3",
    ("Evt16", "Sta12"): "AA-3",
    ("Evt16", "Sta13"): "AA-2",
    ("Evt17", "Sta2"): "AA-5",
    ("Evt17", "Sta3"): "AA-4",
    ("Evt17", "Sta4"): "AA-4",
    ("Evt17", "Sta5"): "AA-4",
    ("Evt17", "Sta6"): "AA-4",
    ("Evt17", "Sta7"): "AA-4",
    ("Evt17", "Sta8"): "AA-4",
    ("Evt17", "Sta9"): "AA-4",
    ("Evt17", "Sta10"): "AA-4",
    ("Evt17", "Sta11"): "AA-4",
    ("Evt17", "Sta12"): "AA-4",
    ("Evt17", "Sta13"): "AR-5",
    ("Evt18", "Sta2"): "AA-2",
    ("Evt18", "Sta13"): "AA-2",
    ("Evt19", "Sta2"): "AA-1",
    ("Evt19", "Sta3"): "AA-8",
    ("Evt19", "Sta5"): "AA-8",
    ("Evt19", "Sta6"): "AA-8",
    ("Evt19", "Sta7"): "AA-8",
    ("Evt19", "Sta8"): "AA-8",
    ("Evt19", "Sta9"): "AA-8",
    ("Evt19", "Sta10"): "AA-8",
    ("Evt19", "Sta11"): "AA-8",
    ("Evt19", "Sta12"): "AA-8",
    ("Evt19", "Sta13"): "AA-7",
}



"""
Implementation of the service parameter primitives.
"""
import logging
from typing import Optional, List, Any, Union, Tuple, cast, Type

from pydicom.uid import UID

from pynetdicom._globals import OptionalUIDType
from pynetdicom.pdu_items import (
    MaximumLengthSubItem,
    ImplementationClassUIDSubItem,
    ImplementationVersionNameSubItem,
    AsynchronousOperationsWindowSubItem,
    SCP_SCU_RoleSelectionSubItem,
    SOPClassExtendedNegotiationSubItem,
    SOPClassCommonExtendedNegotiationSubItem,
    UserIdentitySubItemRQ,
    UserIdentitySubItemAC,
)
from pynetdicom.presentation import PresentationContext
from pynetdicom.utils import validate_uid, decode_bytes, set_ae, set_uid
from pynetdicom._globals import DEFAULT_MAX_LENGTH

LOGGER = logging.getLogger("pynetdicom.pdu_primitives")

_PDUPrimitiveType = Union["A_ASSOCIATE", "A_RELEASE", "A_ABORT", "A_P_ABORT", "P_DATA"]
_UserInformationPrimitiveType = List[
    Union[
        "MaximumLengthNotification",
        "ImplementationClassUIDNotification",
        "ImplementationVersionNameNotification",
        "AsynchronousOperationsWindowNegotiation",
        "SCP_SCU_RoleSelectionNegotiation",
        "SOPClassExtendedNegotiation",
        "UserIdentityNegotiation",
        "SOPClassCommonExtendedNegotiation",
    ]
]
_UI = Union[
    "MaximumLengthNotification",
    "ImplementationClassUIDNotification",
    "ImplementationVersionNameNotification",
    "AsynchronousOperationsWindowNegotiation",
    "SCP_SCU_RoleSelectionNegotiation",
    "SOPClassExtendedNegotiation",
    "SOPClassCommonExtendedNegotiation",
    "UserIdentityNegotiation",
]
_UITypes = Union[
    Type["MaximumLengthNotification"],
    Type["ImplementationClassUIDNotification"],
    Type["ImplementationVersionNameNotification"],
    Type["AsynchronousOperationsWindowNegotiation"],
    Type["SCP_SCU_RoleSelectionNegotiation"],
    Type["SOPClassExtendedNegotiation"],
    Type["SOPClassCommonExtendedNegotiation"],
    Type["UserIdentityNegotiation"],
]


# TODO: Rename to UserInformation
class ServiceParameter:
    """Base class for Service Parameters"""

    def __eq__(self, other: Any) -> bool:
        """Equality of two ServiceParameters"""
        if isinstance(other, self.__class__):
            return other.__dict__ == self.__dict__

        return False

    def __ne__(self, other: Any) -> bool:
        """Inequality of two ServiceParameters"""
        return not self == other


# Association Service primitives
class A_ASSOCIATE:
    """An A-ASSOCIATE primitive.

    The establishment of an association between two AEs shall be performed
    through ACSE A-ASSOCIATE request, indication, response and confirmation
    primitives.

    The initiator of the service is called the Requestor and the user that
    receives the request is the Acceptor.

    The A-ASSOCIATE primitive is used by the DUL provider to send/receive
    information about the association. It gets converted to
    A-ASSOCIATE-RQ, -AC, -RJ PDUs that are sent to the peer DUL provider and
    gets deconverted from -RQ, -AC, -RJ PDUs received from the peer.

    +------------------+---------+------------+----------+--------------+
    | Parameter        | Request | Indication | Response | Confirmation |
    +------------------+---------+------------+----------+--------------+
    | app context name | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | calling ae title | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | called ae title  | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | user info        | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | result           |         |            | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | source           |         |            |          | M            |
    +------------------+---------+------------+----------+--------------+
    | diagnostic       |         |            | U        | C(=)         |
    +------------------+---------+------------+----------+--------------+
    | calling pres add | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | called pres add  | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | pres contxt list | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | pres list result |         |            | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | mode             | UF      | MF(=)      |          |              |
    +------------------+---------+------------+----------+--------------+
    | resp ae title    |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+
    | resp pres add    |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+
    | pres/sess req    | UF      | UF(=)      | UF       | UF(=)        |
    +------------------+---------+------------+----------+--------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    References
    ----------

    * DICOM Standard, Part 8,
      :dcm:`Section 7.1.1<part08/chapter_7.html#sect_7.1.1>`
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self) -> None:
        self._application_context_name: Optional[UID] = None
        self._calling_ae_title: str = "DEFAULT"
        self._called_ae_title: str = "DEFAULT"
        self._user_information: _UserInformationPrimitiveType = []
        self._result: Optional[int] = None
        self._result_source: Optional[int] = None
        self._diagnostic: Optional[int] = None
        self._calling_presentation_address: Optional[Tuple[str, int]] = None
        self._called_presentation_address: Optional[Tuple[str, int]] = None
        self._presentation_context_definition_list: List[PresentationContext] = []
        self._presentation_context_definition_results_list: List[
            PresentationContext
        ] = []

    @property
    def application_context_name(self) -> Optional[UID]:
        """Get or set the *Application Context Name* parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The application context name proposed by the *Requestor*.
            *Acceptor* returns either the same or a different name. Returned
            name specifies the application context used for the association.
            See the DICOM Standard, Part 8, :dcm:`Annex A
            <part08/chapter_A.html>`. The application context name shall
            be a valid UID or UID string and for version 3 of the DICOM
            Standard should be ``'1.2.840.10008.3.1.1.1'``

        Returns
        -------
        pydicom.uid.UID
        """
        return self._application_context_name

    @application_context_name.setter
    def application_context_name(self, value: OptionalUIDType) -> None:
        """Set the Application Context Name parameter."""
        self._application_context_name = set_uid(value, "Application Context Name")

    @property
    def called_ae_title(self) -> str:
        """Get or set the *Called AE Title* parameter.

        Parameters
        ----------
        value : str or bytes
            The Called AE Title as a string or bytes object. Cannot be an empty
            string and will be truncated to 16 characters long

        Returns
        -------
        bytes
        """
        return self._called_ae_title

    @called_ae_title.setter
    def called_ae_title(self, value: str) -> None:
        """Set the Called AE Title parameter."""
        self._called_ae_title = cast(
            str, set_ae(value, "Called AE Title", False, False)
        )

    @property
    def called_presentation_address(self) -> Optional[Tuple[str, int]]:
        """Get or set the *Called Presentation Address* parameter.

        Parameters
        ----------
        value : (str, int) tuple
            A tuple containing a valid TCP/IP address string and the port
            number as an int
        """
        return self._called_presentation_address

    @called_presentation_address.setter
    def called_presentation_address(self, value: Optional[Tuple[str, int]]) -> None:
        """Set the Called Presentation Address parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, tuple):
            if (
                len(value) == 2
                and isinstance(value[0], str)
                and isinstance(value[1], int)
            ):
                self._called_presentation_address = value
            else:
                LOGGER.error(
                    "A_ASSOCIATE.called_presentation_address must "
                    "be (str, int) tuple"
                )
                raise TypeError(
                    "A_ASSOCIATE.called_presentation_address "
                    "must be (str, int) tuple"
                )
        elif value is None:
            self._called_presentation_address = value
        else:
            LOGGER.error(
                "A_ASSOCIATE.called_presentation_address must be (str, int) tuple"
            )
            raise TypeError(
                "A_ASSOCIATE.called_presentation_address must be (str, int) tuple"
            )

    @property
    def calling_ae_title(self) -> str:
        """Get or set the *Calling AE Title* parameter.

        Parameters
        ----------
        value : str or bytes
            The Calling AE Title as a string or bytes object. Cannot be an
            empty string and will be truncated to 16 characters long.

        Returns
        -------
        bytes
        """
        return self._calling_ae_title

    @calling_ae_title.setter
    def calling_ae_title(self, value: str) -> None:
        """Set the Calling AE Title parameter."""
        self._calling_ae_title = cast(
            str, set_ae(value, "Calling AE Title", False, False)
        )

    @property
    def calling_presentation_address(self) -> Optional[Tuple[str, int]]:
        """Get or set the *Calling Presentation Address* parameter.

        Parameters
        ----------
        value : (str, int) tuple
            A tuple containing a valid TCP/IP address string and the port
            number as an int
        """
        return self._calling_presentation_address

    @calling_presentation_address.setter
    def calling_presentation_address(self, value: Optional[Tuple[str, int]]) -> None:
        """Set the A-ASSOCIATE Service primitive's Calling Presentation
        Address parameter.
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, tuple):
            if (
                len(value) == 2
                and isinstance(value[0], str)
                and isinstance(value[1], int)
            ):
                self._calling_presentation_address = value
            else:
                LOGGER.error(
                    "A_ASSOCIATE.calling_presentation_address must "
                    "be (str, int) tuple"
                )
                raise TypeError(
                    "A_ASSOCIATE.calling_presentation_address "
                    "must be (str, int) tuple"
                )
        elif value is None:
            self._calling_presentation_address = value
        else:
            LOGGER.error(
                "A_ASSOCIATE.calling_presentation_address must be (str, int) tuple"
            )
            raise TypeError(
                "A_ASSOCIATE.calling_presentation_address must be (str, int) tuple"
            )

    @property
    def diagnostic(self) -> Optional[int]:
        """Get or set the *Diagnostic* parameter.

        Parameters
        ----------
        value : int
            If `result_source` is "UL service-user" then allowed values are:

            * 1: no reason given
            * 2: application context name not supported
            * 3: calling AE title not recognised
            * 7: called AE title not recognised

            If `result_source` is "UL service-provider (ACSE related function)"
            then allowed values are:

            * 1: no reason given
            * 2: protocol version not supported"

            If `result_source` is "UL service-provider (Presentation related
            function)" then allowed values are:

            * 1: temporary congestion
            * 2: local limit exceeded
        """
        return self._diagnostic

    @diagnostic.setter
    def diagnostic(self, value: Optional[int]) -> None:
        """
        Set the A-ASSOCIATE Service primitive's Diagnostic parameter."""
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [1, 2, 3, 7]:
            LOGGER.error("A_ASSOCIATE.diagnostic set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.diagnostic value")

        self._diagnostic = value

    @property
    def implementation_class_uid(self) -> Optional[UID]:
        """Get or set the *Implementation Class UID* as
        :class:`~pydicom.uid.UID`.

        If the A_ASSOCIATE.user_information list contains an
        ImplementationClassUIDNotification item then set its
        implementation_class_uid value. If not then add a
        ImplementationClassUIDNotification item and set its
        implementation_class_uid value.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The value for the Implementation Class UID
        """
        for item in self.user_information:
            if isinstance(item, ImplementationClassUIDNotification):
                if item.implementation_class_uid is None:
                    LOGGER.error("Implementation Class UID has not been set")
                    raise ValueError("Implementation Class UID has not been set")

                return item.implementation_class_uid

        LOGGER.error("Implementation Class UID has not been set")
        raise ValueError("Implementation Class UID has not been set")

    @implementation_class_uid.setter
    def implementation_class_uid(self, value: OptionalUIDType) -> None:
        """Set the Implementation Class UID."""
        # Type and value checking for the implementation_class_uid parameter is
        #   done by the ImplementationClassUIDNotification class

        # Check for a ImplementationClassUIDNotification item
        found_item = False
        for item in self.user_information:
            if isinstance(item, ImplementationClassUIDNotification):
                found_item = True
                item.implementation_class_uid = value  # type: ignore

        # No ImplementationClassUIDNotification item found
        if not found_item:
            imp_uid = ImplementationClassUIDNotification()
            imp_uid.implementation_class_uid = value  # type: ignore
            self.user_information.append(imp_uid)

    @property
    def maximum_length_received(self) -> Optional[int]:
        """Get or set the *Maximum Length Received* as :class:`int`.

        If the A_ASSOCIATE.user_information list contains a
        MaximumLengthNotification item then set its maximum_length_received
        value. If not then add a MaximumLengthNotification item and set its
        maximum_length_received value.

        Parameters
        ----------
        value : int
            The maximum length of each P-DATA in bytes
        """
        for item in self.user_information:
            if isinstance(item, MaximumLengthNotification):
                return item.maximum_length_received

        return None

    @maximum_length_received.setter
    def maximum_length_received(self, value: int) -> None:
        """Set the Maximum Length Received."""
        # Type and value checking for the maximum_length_received parameter is
        #   done by the MaximumLengthNotification class

        # Check for a MaximumLengthNotification item
        found_item = False

        for item in self.user_information:
            if isinstance(item, MaximumLengthNotification):
                found_item = True
                item.maximum_length_received = value

        # No MaximumLengthNotification item found
        if not found_item:
            max_length = MaximumLengthNotification()
            max_length.maximum_length_received = value
            self.user_information.append(max_length)

    @property
    def mode(self) -> str:
        """Return the *Mode* parameter as :class:`str`."""
        return "normal"

    @property
    def presentation_context_definition_list(self) -> List[PresentationContext]:
        """Get or set the *Presentation Context Definition List*.

        Parameters
        ----------
        value_list : list of presentation.PresentationContext
            The Presentation Contexts proposed by the Association Requestor
        """
        return self._presentation_context_definition_list

    @presentation_context_definition_list.setter
    def presentation_context_definition_list(
        self, value_list: List[PresentationContext]
    ) -> None:
        """Set the A-ASSOCIATE Service primitive's Presentation Context
        Definition List parameter.
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            valid_items = []
            for item in value_list:
                if isinstance(item, PresentationContext):
                    valid_items.append(item)
                else:
                    LOGGER.warning(
                        "Attempted to set "
                        "A_ASSOCIATE.presentation_context_definition_list to "
                        "a list which includes an invalid items"
                    )

            self._presentation_context_definition_list = valid_items

        else:
            LOGGER.error(
                "A_ASSOCIATE.presentation_context_definition_list must be a list"
            )
            raise TypeError(
                "A_ASSOCIATE.presentation_context_definition_list must be a list"
            )

    @property
    def presentation_context_definition_results_list(self) -> List[PresentationContext]:
        """Get or set the *Presentation Context Definition Results List*.

        Parameters
        ----------
        value_list : list of presentation.PresentationContext
            The results of the Presentation Contexts proposal by the
            Association Requestor
        """
        return self._presentation_context_definition_results_list

    @presentation_context_definition_results_list.setter
    def presentation_context_definition_results_list(
        self, value_list: List[PresentationContext]
    ) -> None:
        """Set the Presentation Context Definition Results List parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            valid_items = []
            for item in value_list:
                if isinstance(item, PresentationContext):
                    valid_items.append(item)
                else:
                    LOGGER.warning(
                        "Attempted to set A_ASSOCIATE.presentation"
                        "_context_definition_results_list to a "
                        "list which includes one or more invalid "
                        "items"
                    )

            self._presentation_context_definition_results_list = valid_items

        else:
            LOGGER.error(
                "A_ASSOCIATE.presentation_context_definition_"
                "results_list must be a list"
            )
            raise TypeError(
                "A_ASSOCIATE.presentation_context_definition_"
                "results_list must be a list"
            )

    @property
    def presentation_requirements(self) -> str:
        """Return the *Presentation Kernel* as :class:`str`."""
        return "Presentation Kernel"

    @property
    def reason_str(self) -> str:
        """Return the rejection reason as str."""
        reasons = {
            1: {
                1: "No reason given",
                2: "Application context name not supported",
                3: "Calling AE title not recognised",
                4: "Reserved",
                5: "Reserved",
                6: "Reserved",
                7: "Called AE title not recognised",
                8: "Reserved",
                9: "Reserved",
                10: "Reserved",
            },
            2: {1: "No reason given", 2: "Protocol version not supported"},
            3: {
                0: "Reserved",
                1: "Temporary congestion",
                2: "Local limit exceeded",
                3: "Reserved",
                4: "Reserved",
                5: "Reserved",
                6: "Reserved",
                7: "Reserved",
            },
        }
        result = cast(int, self.result_source)
        diagnostic = cast(int, self.diagnostic)
        try:
            return reasons[result][diagnostic]
        except KeyError:
            LOGGER.warning(
                f"Invalid A-ASSOCIATE 'Result Source' {result} and/or "
                f"'Diagnostic' {diagnostic} values"
            )
            return "(no value available)"

    @property
    def responding_ae_title(self) -> Optional[str]:
        """Return the *Responding AE Title* parameter."""
        return self.called_ae_title

    @property
    def responding_presentation_address(self) -> Optional[Tuple[str, int]]:
        """Get the Responding Presentation Address parameter."""
        return self.called_presentation_address

    @property
    def result(self) -> Optional[int]:
        """Get or set the *Result* parameter.

        Parameters
        ----------
        value : int
            One of the following:

            * 0: accepted
            * 1: rejected (permanent)
            * 2: rejected (transient)

        Returns
        -------
        int
        """
        return self._result

    @result.setter
    def result(self, value: Optional[int]) -> None:
        """Set the A-ASSOCIATE Service primitive's Result parameter."""
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [0, 1, 2]:
            LOGGER.error("A_ASSOCIATE.result set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.result value")

        self._result = value

    @property
    def result_source(self) -> Optional[int]:
        """Get or set the *Result Source* parameter.

        Parameters
        ----------
        value : int
            One of the following:

            * 1: UL service-user
            * 2: UL service-provider (ACSE related function)
            * 3: UL service-provider (presentation related function)

        Returns
        -------
        int
        """
        return self._result_source

    @result_source.setter
    def result_source(self, value: Optional[int]) -> None:
        """Set the A-ASSOCIATE Service primitive's Result Source parameter."""
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [1, 2, 3]:
            LOGGER.error("A_ASSOCIATE.result_source set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.result_source value")

        self._result_source = value

    @property
    def result_str(self) -> str:
        """Return the result as str."""
        results = {0: "Accepted", 1: "Rejected Permanent", 2: "Rejected Transient"}

        if self.result not in results:
            LOGGER.warning(f"Invalid A-ASSOCIATE 'Result' {self.result}")
            return "(no value available)"

        return results[self.result]

    @property
    def session_requirements(self) -> str:
        """Return the *Session Requirements* as :class:`str`."""
        return ""

    @property
    def source_str(self) -> str:
        """Return the reject source as str."""
        sources = {
            1: "Service User",
            2: "Service Provider (ACSE)",
            3: "Service Provider (Presentation)",
        }
        if self.result_source not in sources:
            LOGGER.warning(f"Invalid A-ASSOCIATE 'Result Source' {self.result_source}")
            return "(no value available)"

        return sources[self.result_source]

    @property
    def user_information(self) -> _UserInformationPrimitiveType:
        """Get or set the *User Information* parameter.

        Parameters
        ----------
        value_list : list of user information class objects
            A list of user information objects, must contain at least
            MaximumLengthNotification and ImplementationClassUIDNotification
        """
        return self._user_information

    @user_information.setter
    def user_information(self, value_list: _UserInformationPrimitiveType) -> None:
        """Set the A-ASSOCIATE primitive's User Information parameter."""
        # pylint: disable=attribute-defined-outside-init
        valid_usr_info_items = []

        if isinstance(value_list, list):
            # Iterate through the items and check they're an acceptable class
            for item in value_list:
                if item.__class__.__name__ in [
                    "MaximumLengthNotification",
                    "ImplementationClassUIDNotification",
                    "ImplementationVersionNameNotification",
                    "AsynchronousOperationsWindowNegotiation",
                    "SCP_SCU_RoleSelectionNegotiation",
                    "SOPClassExtendedNegotiation",
                    "SOPClassCommonExtendedNegotiation",
                    "UserIdentityNegotiation",
                ]:
                    valid_usr_info_items.append(item)
                else:
                    LOGGER.info(
                        "Attempted to set "
                        "A_ASSOCIATE.user_information to a list "
                        "which includes an unsupported item"
                    )
        else:
            LOGGER.error("A_ASSOCIATE.user_information must be a list")
            raise TypeError("A_ASSOCIATE.user_information must be a list")

        self._user_information = valid_usr_info_items


class A_RELEASE:
    """An A-RELEASE primitive.

    +------------------+---------+------------+----------+--------------+
    | Parameter        | Request | Indication | Response | Confirmation |
    +------------------+---------+------------+----------+--------------+
    | reason           | UF      | UF(=)      | UF       | UF(=)        |
    +------------------+---------+------------+----------+--------------+
    | user info        |NU       | NU(=)      | NU       | NU(=)        |
    +------------------+---------+------------+----------+--------------+
    | result           |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    References
    ----------
    * DICOM Standard, Part 8, :dcm:`Section 7.2<part08/sect_7.2.html>`
    """

    def __init__(self) -> None:
        self._result: Optional[str] = None

    @property
    def reason(self) -> str:
        """Return the *Reason* parameter."""
        return "normal"

    @property
    def result(self) -> Optional[str]:
        """Get or set the *Result* parameter.

        Parameters
        ----------
        value : str
            Must be ``None`` for request and indication, ``'affirmative'``
            for response and confirmation.
        """
        return self._result

    @result.setter
    def result(self, value: Optional[str]) -> None:
        """Set the Result parameter."""
        # pylint: disable=attribute-defined-outside-init
        if value is not None and value != "affirmative":
            LOGGER.error("A_RELEASE.result must be None or 'affirmative'")
            raise ValueError("A_RELEASE.result must be None or 'affirmative'")

        self._result = value


class A_ABORT:
    """An A-ABORT primitive.

    +------------------+---------+------------+
    | Parameter        | Request | Indication |
    +------------------+---------+------------+
    | abort source     |         | M          |
    +------------------+---------+------------+
    | user info        |NU       | NU(=)      |
    +------------------+---------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    References
    ----------

    * DICOM Standard, Part 8, :dcm:`Section 7.3<part08/sect_7.3.html>`
    """

    def __init__(self) -> None:
        self._abort_source: Optional[int] = None

    @property
    def abort_source(self) -> Optional[int]:
        """Get or set the *Abort Source*.

        Parameters
        ----------
        value : int or None
            The abort source, one of:

            * ``0``: Upper layer service user
            * ``1``: Reserved
            * ``2``: Upper layer service provider
        """
        if self._abort_source is None:
            LOGGER.error("A_ABORT.abort_source value not set")
            raise ValueError("A_ABORT.abort_source value not set")

        return self._abort_source

    @abort_source.setter
    def abort_source(self, value: Optional[int]) -> None:
        """Set the Abort Source."""
        if value in [0, 1, 2, None]:
            self._abort_source = value
        else:
            msg = f"Invalid A-ABORT 'Source' value '{value}'"
            LOGGER.error(msg)
            raise ValueError(msg)


class A_P_ABORT:
    """
    An A-P-ABORT primitive.

    +------------------+------------+
    | Parameter        | Indication |
    +------------------+------------+
    | provider reason  | P          |
    +------------------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | P   - Provider initiated
    | (=) - shall have same value as request or response

    References
    ----------

    * DICOM Standard, Part 8, :dcm:`Section 7.4<part08/sect_7.4.html>`
    """

    def __init__(self) -> None:
        self._provider_reason: Optional[int] = None

    @property
    def provider_reason(self) -> Optional[int]:
        """Return the *Provider Reason*.

        Parameters
        ----------
        value : int
            Indicates the reason for the abort. Allowed values are:

            * ``0``: reason not specified
            * ``1``: unrecognised PDU
            * ``2``: unexpected PDU
            * ``4``: unrecognised PDU parameter
            * ``5``: unexpected PDU parameter
            * ``6``: invalid PDU parameter value
        """
        if self._provider_reason is None:
            LOGGER.error("A_ABORT.provider_reason parameter not set")
            raise ValueError("A_ABORT.provider_reason value not set")

        return self._provider_reason

    @provider_reason.setter
    def provider_reason(self, value: Optional[int]) -> None:
        """Set the Provider Reason."""
        if value in [0, 1, 2, 4, 5, 6, None]:
            self._provider_reason = value
        else:
            msg = "Attempted to set A_P_ABORT.provider_reason to an invalid value"
            LOGGER.error(msg)
            raise ValueError(msg)


class P_DATA:
    """
    A P-DATA primitive.

    +------------------------------+---------+------------+
    | Parameter                    | Request | Indication |
    +------------------------------+---------+------------+
    | presentation data value list | M       | M(=)       |
    +------------------------------+---------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    References
    ----------

    * DICOM Standard, Part 8, :dcm:`Section 7.6<part08/sect_7.6.html>`
    """

    def __init__(self) -> None:
        self._presentation_data_value_list: List[Tuple[int, bytes]] = []

    @property
    def presentation_data_value_list(self) -> List[Tuple[int, bytes]]:
        """Get or set the *Presentation Data Value List*.

        Parameters
        ----------
        presentation_data_value_list : list of [int, bytes]
            Contains one or more Presentation Data Values (PDV), each
            consisting of a Presentation Context ID and User Data values.
            The User Data values are taken from the Abstract Syntax and
            encoded in the Transfer Syntax identified by the Presentation
            Context ID. Each item in the list is ``[Context ID, PDV Data]``
        """
        return self._presentation_data_value_list

    @presentation_data_value_list.setter
    def presentation_data_value_list(self, value_list: List[Tuple[int, bytes]]) -> None:
        """Set the Presentation Data Value List."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            for pdv in value_list:
                if isinstance(pdv, list):
                    if isinstance(pdv[0], int) and isinstance(pdv[1], bytes):
                        pass
                    else:
                        raise TypeError(
                            "P_DATA.presentation_data_value_list "
                            "should be a list of [int, bytes]"
                        )
                else:
                    raise TypeError(
                        "P_DATA.presentation_data_value_list "
                        "should be a list of [ID, PDV]"
                    )
        else:
            raise TypeError(
                "P_DATA.presentation_data_value_list "
                "should be a list of [int, bytes]"
            )

        self._presentation_data_value_list = value_list

    def __str__(self) -> str:
        """String representation of the class."""
        s = "P-DATA\n"
        for pdv in self.presentation_data_value_list:
            header_byte = pdv[1][0]
            s += f"  Context ID: {pdv[0]}\n"
            s += f"  Value Length: {len(pdv[1])} bytes\n"
            s += f"  Message Control Header Byte: {header_byte:08b}\n"

            # xxxxxx01 and xxxxxx011
            if header_byte & 1:
                # xxxxxx11
                if header_byte & 2:
                    s += (
                        "    Command information, last fragment of the "
                        "DIMSE message\n"
                    )
                # xxxxxx01
                else:
                    s += (
                        "    Command information, not the last fragment of "
                        "the DIMSE message\n"
                    )
            # xxxxxx00, xxxxxxx10
            else:
                # xxxxxx10
                if header_byte & 2 != 0:
                    s += (
                        "    Dataset information, last fragment of the "
                        "DIMSE message\n"
                    )
                # xxxxxx00
                else:
                    s += (
                        "    Dataset information, not the last fragment of "
                        "the DIMSE message\n"
                    )

        return s


# User Information Negotiation primitives
class MaximumLengthNotification(ServiceParameter):
    """
    A representation of a Maximum Length Negotiation primitive.

    The maximum length notification allows communicating AEs to limit the size
    of the data for each P-DATA indication. This notification is required for
    all DICOM v3.0 conforming implementations.

    This User Information item is required during association negotiation and
    there must only be a single :class:`MaximumLengthNotification` item.

    References
    ----------

    * DICOM Standard, Part 7,
      :dcm:`Annex D.3.3.1<part07/sect_D.3.3.html#sect_D.3.3.1>`
    * DICOM Standard, Part 8, :dcm:`Annex D.1<part08/chapter_D.html#sect_D.1>`
    """

    def __init__(self) -> None:
        self._maximum_length: int
        self.maximum_length_received = DEFAULT_MAX_LENGTH

    def from_primitive(self) -> MaximumLengthSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.MaximumLengthSubItem
        """
        item = MaximumLengthSubItem()
        item.from_primitive(self)

        return item

    @property
    def maximum_length_received(self) -> int:
        """Get or set the *Maximum Length Received* using :class:`int`.

        Parameters
        ----------
        val : int
            The maximum length of each P-DATA in bytes, must be equal to or
            greater than 0. A value of ``0`` indicates unlimited length
            (``31682`` bytes default).

        Raises
        ------
        ValueError
            If `maximum_length_received` is negative
        TypeError
            If `maximum_length_received` is not an int
        """
        return self._maximum_length

    @maximum_length_received.setter
    def maximum_length_received(self, val: int) -> None:
        """User defined Maximum Length to be used during an Association."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(val, int):
            if val < 0:
                LOGGER.error("Maximum Length Received must be greater than 0")
                raise ValueError("Maximum Length Received must be greater than 0")
            else:
                self._maximum_length = val
        else:
            LOGGER.error("Maximum Length Received must be numerical")
            raise TypeError("Maximum Length Received must be numerical")

    def __str__(self) -> str:
        """String representation of the class."""
        s = [
            "Maximum Length Negotiation",
            f"  Maximum length received: " f"{self.maximum_length_received:d} bytes\n",
        ]
        return "\n".join(s)


class ImplementationClassUIDNotification(ServiceParameter):
    """A representation of a Implementation Class UID Notification primitive.

    The implementation identification notification allows implementations of
    communicating AEs to identify each other at association establishment time.
    It is intended to provider respective and non-ambiguous identification in
    the event of communication problems encountered between two nodes. This
    negotiation is required.

    Implementation identification relies on two pieces of information:

    - Implementation Class UID (required)
    - Implementation Version Name (optional)

    The Implementation Class UID is required during association negotiation and
    there must only be a single :class:`ImplementationClassUIDNotification`
    item.

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     ImplementationClassUIDNotification
    ... )
    >>> item = ImplementationClassUIDNotification()
    >>> item.implementation_class_uid = '1.2.3.4'

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.2<part07/sect_D.3.3.2.html>`
    """

    def __init__(self) -> None:
        self._implementation_class_uid: Optional[UID] = None

    def from_primitive(self) -> ImplementationClassUIDSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.ImplementationClassUIDSubItem

        Raises
        ------
        ValueError
            If no UID is set
        """
        if self.implementation_class_uid is None:
            LOGGER.error(
                "The Implementation Class UID must be set prior to "
                "requesting Association"
            )
            raise ValueError(
                "The Implementation Class UID must be set "
                "prior to requesting Association"
            )

        item = ImplementationClassUIDSubItem()
        item.from_primitive(self)

        return item

    @property
    def implementation_class_uid(self) -> Optional[UID]:
        """Get or set the *Implementation Class UID*.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The value for the Implementation Class UID
        """
        return self._implementation_class_uid

    @implementation_class_uid.setter
    def implementation_class_uid(self, value: OptionalUIDType) -> None:
        """Sets the Implementation Class UID parameter."""
        self._implementation_class_uid = set_uid(value, "Implementation Class UID")

    def __str__(self) -> str:
        """String representation of the class."""
        s = "Implementation Class UID\n"
        s += f"  Implementation class UID: {self.implementation_class_uid}\n"
        return s


class ImplementationVersionNameNotification(ServiceParameter):
    """
    A representation of a Implementation Version Name Notification primitive.

    The implementation identification notification allows implementations of
    communicating AEs to identify each other at Association establishment time.
    It is intended to provider respective and non-ambiguous identification in
    the event of communication problems encountered between two nodes. This
    negotiation is required.

    Implementation identification relies on two pieces of information:

    - Implementation Class UID (required)
    - Implementation Version Name (optional)

    The Implementation Version Name is optional and there may only be a single
    :class:`ImplementationVersionNameNotification` item.

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     ImplementationVersionNameNotification
    ... )
    >>> item = ImplementationVersionNameNotification()
    >>> item.implementation_version_name = b'SOME_NAME'

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.2<part07/sect_D.3.3.2.html>`
    """

    def __init__(self) -> None:
        self._implementation_version_name: Optional[str] = None

    def from_primitive(self) -> ImplementationVersionNameSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.ImplementationVersionNameSubItem

        Raises
        ------
        ValueError
            If no name is set
        """
        if self.implementation_version_name is None:
            raise ValueError(
                "Implementation Version Name must be set prior to Association"
            )

        item = ImplementationVersionNameSubItem()
        item.from_primitive(self)

        return item

    @property
    def implementation_version_name(self) -> Optional[str]:
        """Get or set the *Implementation Version Name*.

        Parameters
        ----------
        value : str or bytes
            The value for the Implementation Version Name

        Raises
        ------
        TypeError
            If `value` is not a str or bytes
        ValueError
            If `value` is empty or longer than 16 characters
        """
        return self._implementation_version_name

    @implementation_version_name.setter
    def implementation_version_name(self, value: Optional[str]) -> None:
        """Sets the Implementation Version Name parameter."""
        self._implementation_version_name = set_ae(
            value, "Implementation Version Name", True, True
        )

    def __str__(self) -> str:
        """String representation of the class."""
        version = self.implementation_version_name
        s = "Implementation Version Name\n"
        s += f"  Implementation version name: {version}\n"

        return s


class AsynchronousOperationsWindowNegotiation(ServiceParameter):
    """
    Representation of the Asynchronous Operations Window Negotiation primitive.

    Allows peer AEs to negotiate the maximum number of outstanding operation
    or sub-operation requests. This negotiation is optional.

    The Asynchronous Operations Window is optional and there may only be a
    single :class:`AsynchronousOperationsWindowNegotiation` item

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC.

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     AsynchronousOperationsWindowNegotiation
    ... )
    >>> item = AsynchronousOperationsWindowNegotiation()
    >>> item.maximum_number_operations_invoked = 2
    >>> item.maximum_number_operations_performed = 1

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.3<part07/sect_D.3.3.3.html>`
    """

    def __init__(self) -> None:
        self._maximum_number_operations_invoked = 1
        self._maximum_number_operations_performed = 1

    def from_primitive(self) -> AsynchronousOperationsWindowSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.AsynchronousOperationsWindowSubItem
        """
        item = AsynchronousOperationsWindowSubItem()
        item.from_primitive(self)

        return item

    @property
    def maximum_number_operations_invoked(self) -> int:
        """Get or set the *Maximum Number Operations Invoked*.

        Parameters
        ----------
        value : int
            The maximum number of operations invoked

        Raises
        ------
        TypeError
            If `value` is not an int
        ValueError
            If `value` is less than 0
        """
        return self._maximum_number_operations_invoked

    @maximum_number_operations_invoked.setter
    def maximum_number_operations_invoked(self, value: int) -> None:
        """Sets the Maximum Number Operations Invoked parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, int):
            pass
        else:
            LOGGER.error("Maximum Number Operations Invoked must be an int")
            raise TypeError("Maximum Number Operations Invoked must be an int")

        if value < 0:
            raise ValueError("Maximum Number Operations Invoked must be greater than 0")

        self._maximum_number_operations_invoked = value

    @property
    def maximum_number_operations_performed(self) -> int:
        """Get or set the *Maximum Number Operations Performed*.

        Parameters
        ----------
        value : int
            The maximum number of operations performed

        Raises
        ------
        TypeError
            If `value` is not an int
        ValueError
            If `value` is less than 0
        """
        return self._maximum_number_operations_performed

    @maximum_number_operations_performed.setter
    def maximum_number_operations_performed(self, value: int) -> None:
        """Sets the Maximum Number Operations Performed parameter"""
        # pylint: disable=attribute-defined-outside-init
        if not isinstance(value, int):
            LOGGER.error("Maximum Number Operations Performed must be an int")
            raise TypeError("Maximum Number Operations Performed must be an int")

        if value < 0:
            raise ValueError(
                "Maximum Number Operations Performed must be greater than 0"
            )

        self._maximum_number_operations_performed = value

    def __str__(self) -> str:
        """String representation of the class."""
        invoked = self.maximum_number_operations_invoked
        performed = self.maximum_number_operations_performed
        s = "Asynchronous Operations Window\n"
        s += f"  Maximum number operations invoked: {invoked:d}\n"
        s += f"  Maximum number operations performed: {performed:d}\n"

        return s


class SCP_SCU_RoleSelectionNegotiation(ServiceParameter):
    """
    A representation of the SCP/SCU Role Selection Negotiation primitive.

    Allows peer AEs to negotiate the roles in which they will serve for each
    SOP Class or Meta SOP Class supported on the association. This negotiation
    is optional.

    The association *Requestor* may use one SCP/SCU Role Selection item for
    each SOP Class as identified by its corresponding Abstract Syntax Name and
    shall be one of three role values:

    - *Requestor* is SCU only
    - *Requestor* is SCP only
    - *Requestor* is both SCU/SCP

    If the SCP/SCU Role Selection item is absent the default role for a
    *Requestor* is SCU and for an *Acceptor* is SCP.

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC.

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation
    >>> item = SCP_SCU_RoleSelectionNegotiation()
    >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.1.2'
    >>> item.scu_role = True
    >>> item.scp_role = False

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.4<part07/sect_D.3.3.4.html>`
    """

    def __init__(self) -> None:
        self._sop_class_uid: Optional[UID] = None
        self._scu_role: Optional[bool] = None
        self._scp_role: Optional[bool] = None

    def from_primitive(self) -> SCP_SCU_RoleSelectionSubItem:
        """Convert the primitive to a PDU item ready to be encoded

        Returns
        -------
        item : pdu_items.SCP_SCU_RoleSelectionSubItem

        Raises
        ------
        ValueError
            If no SOP Class UID, SCU Role or SCP Role is set
        ValueError
            If SCU Role and SCP Role are both False
        """
        if self.sop_class_uid is None:
            LOGGER.error("'sop_class_uid' must be set prior to Association")
            raise ValueError("'sop_class_uid' must be set prior to Association")

        # To get to this point self.sop_class_uid must be set
        if not self.scu_role and not self.scp_role:
            uid = self.sop_class_uid
            msg = f"SCU and SCP Roles cannot both be unsupported for '{uid}'"
            LOGGER.error(msg)
            raise ValueError(msg)

        item = SCP_SCU_RoleSelectionSubItem()
        item.from_primitive(self)

        return item

    @property
    def scp_role(self) -> Optional[bool]:
        """Get or set the *SCP Role*.

        Parameters
        ----------
        value : bool
            True if supported, False otherwise (default)

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        return self._scp_role

    @scp_role.setter
    def scp_role(self, value: Optional[bool]) -> None:
        """Sets the SCP Role parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool) or value is None:
            pass
        else:
            LOGGER.error("SCP Role must be boolean")
            raise TypeError("SCP Role must be boolean")

        self._scp_role = value

    @property
    def scu_role(self) -> Optional[bool]:
        """Get or set the *SCU Role*.

        Parameters
        ----------
        value : bool
            True if supported, False otherwise

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        return self._scu_role

    @scu_role.setter
    def scu_role(self, value: Optional[bool]) -> None:
        """Sets the SCU Role parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool) or value is None:
            pass
        else:
            LOGGER.error("SCU Role must be boolean")
            raise TypeError("SCU Role must be boolean")

        self._scu_role = value

    @property
    def sop_class_uid(self) -> Optional[UID]:
        """Get or set the *SOP Class UID*.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Abstract Syntax UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value: OptionalUIDType) -> None:
        """Sets the SOP Class UID parameter."""
        self._sop_class_uid = set_uid(value, "SOP Class UID")


class SOPClassExtendedNegotiation(ServiceParameter):
    r"""
    A representation of the SOP Class Extended Negotiation primitive.

    Allows peer AEs to exchange application information defined by specific
    Service Class specifications. Each Service Class is required to document
    the application information it supports and how this information is
    negotiated between SCUs and SCPs.

    The SOP Class Extended Negotiation is optional and there may only be a
    single :class:`SOPClassExtendedNegotiation` item for each available SOP
    Class UID.

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Examples
    --------

    .. code-block:: python

       >>> from pynetdicom.pdu_primitives import SOPClassExtendedNegotiation
       >>> item = SOPClassExtendedNegotiation()
       >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.2.1.3'
       >>> item.service_class_application_information = b'\x01'

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.5<part07/sect_D.3.3.5.html>`
    """

    def __init__(self) -> None:
        self._sop_class_uid: Optional[UID] = None
        self._service_class_application_information: Optional[bytes] = None

    def from_primitive(self) -> SOPClassExtendedNegotiationSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.SOPClassExtendedNegotiationSubItem

        Raises
        ------
        ValueError
            If `sop_class_uid` or `service_class_application_information` are
            not set
        """
        if (
            self.sop_class_uid is None
            or self.service_class_application_information is None
        ):
            LOGGER.error(
                "SOP Class UID and Service Class Application "
                "Information must be set prior to Association "
                "negotiation"
            )
            raise ValueError(
                "SOP Class UID and Service Class Application "
                "Information must be set prior to Association "
                "negotiation"
            )

        item = SOPClassExtendedNegotiationSubItem()
        item.from_primitive(self)

        return item

    @property
    def service_class_application_information(self) -> Optional[bytes]:
        """Get or set the *Service Class Application Information*.

        Parameters
        ----------
        value : bytes
            The Service Class Application Information as per the Service Class
            Specifications (see PS3.4)

        Raises
        ------
        TypeError
            If `value` is not a bytes object.
        """
        return self._service_class_application_information

    @service_class_application_information.setter
    def service_class_application_information(self, value: Optional[bytes]) -> None:
        """Sets the Service Class Application Information parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes) or value is None:
            pass
        else:
            LOGGER.error(
                "Service Class Application Information should be a bytes object"
            )
            raise TypeError(
                "Service Class Application Information should be a bytes object"
            )

        self._service_class_application_information = value

    @property
    def sop_class_uid(self) -> Optional[UID]:
        """Get or set the *SOP Class UID*.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Abstract Syntax UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value: OptionalUIDType) -> None:
        """Sets the SOP Class UID parameter."""
        self._sop_class_uid = set_uid(value, "SOP Class UID")


class SOPClassCommonExtendedNegotiation(ServiceParameter):
    """
    A representation of the SOP Class Common Extended Negotiation primitive.

    Allows peer AEs to exchange generic application information.

    The SOP Class Common Extended Negotiation is optional and there may only be
    a single :class:`SOPClassCommonExtendedNegotiation` item for each available
    SOP Class UID.

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import SOPClassCommonExtendedNegotiation
    >>> item = SOPClassCommonExtendedNegotiation()
    >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.1.88.40'
    >>> item.service_class_uid = '1.2.840.10008.4.2'
    >>> item.related_general_sop_class_identification = [
    ...     '1.2.840.10008.5.1.4.1.1.88.22'
    ... ]

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.6<part07/sect_D.3.3.6.html>`
    """

    def __init__(self) -> None:
        self._service_class_uid: Optional[UID] = None
        self._sop_class_uid: Optional[UID] = None
        self._related_general_sop_class_identification: List[UID] = []

    def from_primitive(self) -> SOPClassCommonExtendedNegotiationSubItem:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.SOPClassCommonExtendedNegotiationSubItem

        Raises
        ------
        ValueError
            If `sop_class_uid` or `service_class_uid` are not set
        """
        if self.sop_class_uid is None or self.service_class_uid is None:
            LOGGER.error(
                "SOP Class UID and Service Class UID must be set "
                "prior to Association negotiation"
            )
            raise ValueError(
                "SOP Class UID and Service Class UID must be "
                "set prior to Association negotiation"
            )

        item = SOPClassCommonExtendedNegotiationSubItem()
        item.from_primitive(self)

        return item

    @property
    def related_general_sop_class_identification(self) -> List[UID]:
        """Return the *Related General SOP Class Identification*.

        Parameters
        ----------
        uid_list : list of (pydicom.uid.UID, bytes or str)
            A list containing UIDs to be used in the Related General SOP Class
            Identification parameter

        Raises
        ------
        TypeError
            If `uid_list` is not a list
        ValueError
            If `uid_list` contains items that aren't UIDs
        """
        return self._related_general_sop_class_identification

    @related_general_sop_class_identification.setter
    def related_general_sop_class_identification(
        self, uid_list: List[Union[str, bytes, UID]]
    ) -> None:
        """Sets the Service Class Application Information parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(uid_list, list):
            # Test that all the items in the list are UID compatible and
            #   convert them to pydicom.uid.UID if required
            valid_uid_list = []

            for uid in uid_list:
                if isinstance(uid, UID):
                    pass
                elif isinstance(uid, str):
                    uid = UID(uid)
                elif isinstance(uid, bytes):
                    uid = UID(decode_bytes(uid))
                else:
                    msg = (
                        "Related General SOP Class Identification "
                        "must be a list of pydicom.uid.UID, str or bytes"
                    )
                    LOGGER.error(msg)
                    raise TypeError(msg)

                if uid is not None and not validate_uid(uid):
                    msg = (
                        "Related General SOP Class "
                        "Identification contains an invalid UID"
                    )
                    LOGGER.error(msg)
                    raise ValueError(msg)

                if uid and not uid.is_valid:
                    LOGGER.warning(
                        f"The Related General SOP Class UID '{uid}' is "
                        f"non-conformant"
                    )

                valid_uid_list.append(uid)

            self._related_general_sop_class_identification = valid_uid_list
        else:
            msg = (
                "Related General SOP Class Identification "
                "must be a list of pydicom.uid.UID, str or bytes"
            )
            LOGGER.error(msg)
            raise TypeError(msg)

    @property
    def service_class_uid(self) -> Optional[UID]:
        """Return the *Service Class UID*.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Service Class UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        return self._service_class_uid

    @service_class_uid.setter
    def service_class_uid(self, value: OptionalUIDType) -> None:
        """Sets the Service Class UID parameter."""
        self._service_class_uid = set_uid(value, "Service Class UID")

    @property
    def sop_class_uid(self) -> Optional[UID]:
        """Return the *SOP Class UID*.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The SOP Class UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value: OptionalUIDType) -> None:
        """Sets the SOP Class UID parameter."""
        self._sop_class_uid = set_uid(value, "SOP Class UID")


class UserIdentityNegotiation(ServiceParameter):
    """Representation of the User Identity Negotiation primitive.

    Allows peer AEs to exchange generic application information.

    The SOP Class Common Extended Negotiation is optional and there may only be
    a single :class:`UserIdentityNegotiation` item.

    In general, a User Identity Negotiation request that is accepted will
    result in association establishment and possibly a server response if
    requested and supported by the peer. If a server response is requested but
    not received then the *Requestor* must decide how to proceed.
    An association rejected due to an authorisation failure will be indicated
    using Rejection Permanent with a Source of "DICOM UL service provided (ACSE
    related function)".

    How the *Acceptor* handles authentication is to be implemented by the
    end-user and is outside the scope of the DICOM standard.

    A-ASSOCIATE-RQ

    | `user_identity_type`
    | `positive_response_requested`
    | `primary_field`
    | `secondary_field`

    A-ASSOCIATE-AC
    The `server_response` parameter is required when a response to the User
    Identity Negotiation request is to be issued (although this depends on
    whether or not this is supported by the *Acceptor*).

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import UserIdentityNegotiation
    >>> item = UserIdentityNegotiation()
    >>> item.user_identity_type = 2
    >>> item.positive_response_requested = True
    >>> item.primary_field = b'username'
    >>> item.secondary_field = b'password'

    References
    ----------

    * DICOM Standard, Part 7, :dcm:`Annex D.3.3.7<part07/sect_D.3.3.7.html>`
    """

    def __init__(self) -> None:
        self._user_identity_type: Optional[int] = None
        self._positive_response_requested: bool = False
        self._primary_field: Optional[bytes] = None
        self._secondary_field: Optional[bytes] = b""
        self._server_response: Optional[bytes] = None

    def from_primitive(self) -> Union[UserIdentitySubItemAC, UserIdentitySubItemRQ]:
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.UserIdentitySubItemRQ or
            pdu_items.UserIdentitySubItemAC

        Raises
        ------
        ValueError
            If server_response is None and user_identity_type or primary_field
            are None
        ValueError
            If server_response is None and user_identity_type is 2 and
            secondary_field is None
        """
        # Determine if this primitive is an -RQ or -AC
        item: Union[UserIdentitySubItemRQ, UserIdentitySubItemAC]
        if self.server_response is None:
            # Then an -RQ
            if self.user_identity_type is None or self.primary_field is None:
                msg = (
                    "User Identity Type and Primary Field must be "
                    "set prior to Association negotiation"
                )
                LOGGER.error(msg)
                raise ValueError(msg)

            if self.user_identity_type == 2 and self.secondary_field == b"":
                msg = "Secondary Field must be set when User Identity is 2"
                LOGGER.error(msg)
                raise ValueError(msg)

            item = UserIdentitySubItemRQ()

        else:
            # Then an -AC
            item = UserIdentitySubItemAC()

        item.from_primitive(self)

        return item

    @property
    def positive_response_requested(self) -> bool:
        """Get or set *Positive Response Requested*.

        Parameters
        ----------
        value : bool
            True if response requested, False otherwise

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        return self._positive_response_requested

    @positive_response_requested.setter
    def positive_response_requested(self, value: bool) -> None:
        """Sets the Positive Response Requested parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool):
            pass
        else:
            LOGGER.error("Positive Response Requested must be boolean")
            raise TypeError("Positive Response Requested must be boolean")

        self._positive_response_requested = value

    @property
    def primary_field(self) -> Optional[bytes]:
        """Get or set *Primary Field*.

        Parameters
        ----------
        value : bytes or None
            The username, Kerberos Service ticket or SAML assertion as a
            bytes object.

        Raises
        ------
        TypeError
            If `value` is not bytes or None
        """
        return self._primary_field

    @primary_field.setter
    def primary_field(self, value: Optional[bytes]) -> None:
        """Sets the Primary Field parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error(
                "Primary Field must be bytes if requesting "
                "Association, None otherwise"
            )
            raise TypeError(
                "Primary Field must be bytes if requesting "
                "Association, None otherwise"
            )

        self._primary_field = value

    @property
    def secondary_field(self) -> Optional[bytes]:
        """Get or set the *Secondary Field*.

        Only used when User Identity Type is equal to 2.

        Parameters
        ----------
        value : bytes
            The passcode as a bytes object
        """
        return self._secondary_field

    @secondary_field.setter
    def secondary_field(self, value: Optional[bytes]) -> None:
        """Sets the Secondary Field parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes) or value is None:
            pass
        else:
            LOGGER.error(
                "Secondary Field must be bytes if requesting "
                "Association with User Identity Type equal to 2, "
                "None otherwise"
            )
            raise TypeError(
                "Secondary Field must be bytes if requesting "
                "Association with User Identity Type equal to 2, "
                "None otherwise"
            )

        self._secondary_field = value

    @property
    def server_response(self) -> Optional[bytes]:
        """Get or set the *Server Response*.

        A-ASSOCIATE-AC only.

        Parameters
        ----------
        value : bytes or None
            The server response as a bytes object. The Kerberos Service
            ticket, SAML response or JSON web token if the
            `user_identity_type` is ``3``, ``4`` or ``5``. Shall be
            ``None`` if `user_identity_type` was ``1`` or ``2``.

        Raises
        ------
        TypeError
            If `value` is not bytes or None
        """
        return self._server_response

    @server_response.setter
    def server_response(self, value: Optional[bytes]) -> None:
        """Sets the Server Response parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Server Response must be bytes or None")
            raise TypeError("Server Response must be bytes or None")

        self._server_response = value

    def __str__(self) -> str:
        """String representation of the class."""
        s = "User Identity Parameters\n"
        if self.server_response is None:
            rsp_req = self.positive_response_requested
            s += f"  User identity type: {self.user_identity_type:d}\n"
            s += f"  Positive response requested: {rsp_req}\n"
            s += f"  Primary field: {self.primary_field!r}\n"
            s += f"  Secondary field: {self.secondary_field!r}\n"
        else:
            s += f"  Server response: {self.server_response!r}\n"

        return s

    @property
    def user_identity_type(self) -> Optional[int]:
        """Return the *User Identity Type*.

        Parameters
        ----------
        value : int
            One of the following:

            * 1 - Username as string in UTF-8
            * 2 - Username as string in UTF-8 and passcode
            * 3 - Kerberos Service ticket
            * 4 - SAML Assertion
            * 5 - JSON Web Token

        Raises
        ------
        TypeError
            If `value` is not an int or None
        ValueError
            If `value` is an int and is not 1, 2, 3 or 4
        """
        return self._user_identity_type

    @user_identity_type.setter
    def user_identity_type(self, value: Optional[int]) -> None:
        """Sets the User Identity Type parameter."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, int):
            if value not in [1, 2, 3, 4, 5]:
                LOGGER.error(
                    "User Identity Type must be 1, 2, 3, 4 or 5 if "
                    "requesting Association, None otherwise"
                )
                raise ValueError(
                    "User Identity Type must be 1, 2, 3, 4 or 5 "
                    "if requesting Association, None otherwise"
                )
        elif value is None:
            pass
        else:
            LOGGER.error("User Identity Type must be an int or None")
            raise TypeError("User Identity Type must be an int or None")

        self._user_identity_type = value



"""Database interface for the qrscp application.

Unique Keys
-----------
* At each level one attribute is unique
* A unique key shall uniquely identify a single instance at a given level
* Unique keys **may** be in a C-FIND request's Identifier
* Unique keys **shall** be in a C-MOVE or C-GET request's Identifier
* C-FIND, C-GET and C-MOVE shall support existence and matching of all
* unique keys. All instances managed shall have specific non-zero length
  unique key values

Required Keys
-------------
* Multiple instances may have the same value for required keys.
* Required keys may be in a C-FIND request's Identifier
* Required keys shall not be in a C-GET or C-MOVE request's Identifier
"""

from collections import OrderedDict
import sys

try:
    from sqlalchemy import create_engine, Column, ForeignKey, Integer, String
except ImportError:
    sys.exit("qrscp requires the sqlalchemy package")

from sqlalchemy.ext.declarative import declarative_base

from pydicom.dataset import Dataset

from pynetdicom import build_context
from pynetdicom.sop_class import (
    PatientRootQueryRetrieveInformationModelFind,
    PatientRootQueryRetrieveInformationModelMove,
    PatientRootQueryRetrieveInformationModelGet,
    StudyRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelMove,
    StudyRootQueryRetrieveInformationModelGet,
)


class InvalidIdentifier(Exception):
    pass


# C.2.2.2: The total length of the attribute may be larger than given in Part 5
# C.2.2.2: The VM may be larger than the VM from Part 6, depending
#   on the matching type

# VRs for supported elements - Part 5
# CS - 16 bytes maximum - str
# DA - 8 bytes fixed, format YYYYMMDD - str
# IS - 12 bytes maximum, range - int
# LO - 64 characters maximum - str
# PN - 64 characters maximum per component group (5 components per group) - str
# SH - 16 characters maximum - str
# TM - 14 bytes maximum, format HHMMSS.FFFFFF - str
# UI - 64 bytes maximum - str


# Translate from the element keyword to the db attribute
_TRANSLATION = {
    "PatientID": "patient_id",  # PATIENT | Unique | VM 1 | LO
    "PatientName": "patient_name",  # PATIENT | Required | VM 1 | PN
    "StudyInstanceUID": "study_instance_uid",  # STUDY | Unique | VM 1 | UI
    "StudyDate": "study_date",  # STUDY | Required | VM 1 | DA
    "StudyTime": "study_time",  # STUDY | Required | VM 1 | TM
    "AccessionNumber": "accession_number",  # STUDY | Required | VM 1 | SH
    "StudyID": "study_id",  # STUDY | Required | VM 1 | SH
    "SeriesInstanceUID": "series_instance_uid",  # SERIES | Unique | VM 1 | UI
    "Modality": "modality",  # SERIES | Required | VM 1 | CS
    "SeriesNumber": "series_number",  # SERIES | Required | VM 1 | IS
    "SOPInstanceUID": "sop_instance_uid",  # IMAGE | Unique | VM 1 | UI
    "InstanceNumber": "instance_number",  # IMAGE | Required | VM 1 | IS
}

# Unique and required keys and their level, VR and VM for Patient Root
# Study Root is the same but includes the PATIENT attributes
_ATTRIBUTES = {
    "PatientID": ("PATIENT", "U", "LO", 1),
    "PatientName": ("PATIENT", "R", "PN", 1),
    "StudyInstanceUID": ("STUDY", "U", "UI", 1),
    "StudyDate": ("STUDY", "R", "DA", 1),
    "StudyTime": ("STUDY", "R", "TM", 1),
    "AccessionNumber": ("STUDY", "R", "SH", 1),
    "StudyID": ("STUDY", "R", "SH", 1),
    "SeriesInstanceUID": ("SERIES", "U", "UI", 1),
    "Modality": ("SERIES", "R", "VS", 1),
    "SeriesNumber": ("SERIES", "R", "IS", 1),
    "SOPInstanceUID": ("IMAGE", "U", "UI", 1),
    "InstanceNumber": ("IMAGE", "R", "UI", 1),
}
_PATIENT_ROOT_ATTRIBUTES = OrderedDict(
    {
        "PATIENT": ["PatientID", "PatientName"],
        "STUDY": [
            "StudyInstanceUID",
            "StudyDate",
            "StudyTime",
            "AccessionNumber",
            "StudyID",
        ],
        "SERIES": ["SeriesInstanceUID", "Modality", "SeriesNumber"],
        "IMAGE": ["SOPInstanceUID", "InstanceNumber"],
    }
)
_STUDY_ROOT_ATTRIBUTES = OrderedDict(
    {
        "STUDY": [
            "StudyInstanceUID",
            "StudyDate",
            "StudyTime",
            "AccessionNumber",
            "StudyID",
            "PatientID",
            "PatientName",
        ],
        "SERIES": ["SeriesInstanceUID", "Modality", "SeriesNumber"],
        "IMAGE": ["SOPInstanceUID", "InstanceNumber"],
    }
)

# Supported Information Models
_C_FIND = [
    PatientRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelFind,
]
_C_GET = [
    PatientRootQueryRetrieveInformationModelGet,
    StudyRootQueryRetrieveInformationModelGet,
]
_C_MOVE = [
    PatientRootQueryRetrieveInformationModelMove,
    StudyRootQueryRetrieveInformationModelMove,
]

_PATIENT_ROOT = {
    PatientRootQueryRetrieveInformationModelFind: _PATIENT_ROOT_ATTRIBUTES,
    PatientRootQueryRetrieveInformationModelGet: _PATIENT_ROOT_ATTRIBUTES,
    PatientRootQueryRetrieveInformationModelMove: _PATIENT_ROOT_ATTRIBUTES,
}
_STUDY_ROOT = {
    StudyRootQueryRetrieveInformationModelFind: _STUDY_ROOT_ATTRIBUTES,
    StudyRootQueryRetrieveInformationModelGet: _STUDY_ROOT_ATTRIBUTES,
    StudyRootQueryRetrieveInformationModelMove: _STUDY_ROOT_ATTRIBUTES,
}


def add_instance(ds, session, fpath=None):
    """Add a SOP Instance to the database or update existing instance.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The SOP Instance to be added to the database.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    fpath : str, optional
        The path to where the SOP Instance is stored, taken relative
        to the database file.
    """
    # Check if instance is already in the database
    result = (
        session.query(Instance)
        .filter(Instance.sop_instance_uid == ds.SOPInstanceUID)
        .all()
    )
    if result:
        instance = result[0]
    else:
        instance = Instance()

    # Unique or Required attributes
    required = [
        # (Instance attribute, DICOM keyword, max length, req'd)
        ("patient_id", "PatientID", 16, True),
        ("patient_name", "PatientName", 64, False),
        ("study_instance_uid", "StudyInstanceUID", 64, True),
        ("study_date", "StudyDate", 8, False),
        ("study_time", "StudyTime", 14, False),
        ("accession_number", "AccessionNumber", 16, False),
        ("study_id", "StudyID", 16, False),
        ("series_instance_uid", "SeriesInstanceUID", 64, True),
        ("modality", "Modality", 16, False),
        ("series_number", "SeriesNumber", None, False),
        ("sop_instance_uid", "SOPInstanceUID", 64, True),
        ("instance_number", "InstanceNumber", None, False),
    ]

    # Unique and Required attributes
    for attr, keyword, max_len, unique in required:
        if not unique and keyword not in ds:
            value = None
        else:
            elem = ds[keyword]
            value = elem.value

        if value is not None:
            # All supported attributes have VM 1
            # assert elem.VM == 1
            if max_len:
                if elem.VR == "PN":
                    value = str(value)

                assert len(value) <= max_len
            else:
                assert -(2 ** 31) <= value <= 2 ** 31 - 1

        setattr(instance, attr, value)

    instance.filename = fpath

    # Transfer Syntax UID
    try:
        tsyntax = ds.file_meta.TransferSyntaxUID
        if tsyntax:
            assert len(tsyntax) < 64
            instance.transfer_syntax_uid = tsyntax
    except (AttributeError, AssertionError) as exc:
        pass

    # SOP Class UID
    try:
        uid = ds.SOPClassUID
        if uid:
            assert len(uid) < 64
            instance.sop_class_uid = uid
    except (AttributeError, AssertionError):
        pass

    session.add(instance)
    session.commit()


def build_query(identifier, session, query=None):
    """Perform a query against the database.

    Parameters
    ----------
    identifier : pydicom.dataset.Dataset
        The request's *Identifier* dataset containing the query attributes.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        If not used then start a new query, otherwise extend the existing
        `query`.

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    # VRs for Single Value Matching and Wild Card Matching
    _text_vr = ["AE", "CS", "LO", "LT", "PN", "SH", "ST", "UC", "UR", "UT"]
    for elem in [e for e in identifier if e.keyword in _ATTRIBUTES]:
        vr = elem.VR
        val = elem.value
        # Convert PersonName3 to str
        if vr == "PN" and val:
            val = str(val)

        # Part 4, C.2.2.2.1 Single Value Matching
        if vr != "SQ" and val is not None:
            if vr in _text_vr and ("*" in val or "?" in val):
                pass
            elif vr in ["DA", "TM", "DT"] and "-" in val:
                pass
            else:
                # print('Performing single value matching...')
                query = _search_single_value(elem, session, query)
                continue

        # Part 4, C.2.2.2.3 Universal Matching
        if val is None:
            # print('Performing universal matching...')
            query = _search_universal(elem, session, query)
            continue

        # Part 4, C.2.2.2.2 List of UID Matching
        if vr == "UI":
            # print('Performing list of UID matching...')
            query = _search_uid_list(elem, session, query)
            continue

        # Part 4, C.2.2.2.4 Wild Card Matching
        if vr in _text_vr and ("*" in val or "?" in val):
            # print('Performing wildcard matching...')
            query = _search_wildcard(elem, session, query)
            continue

        # Part 4, C.2.2.2.5 Range Matching
        if vr in ["DT", "TM", "DA"] and "-" in val:
            query = _search_range(elem, session, query)
            continue

        # Part 4, C.2.2.2.6 Sequence Matching
        #   No supported attributes are sequences

    return query


def _check_identifier(identifier, model):
    """Check that the C-FIND, C-GET or C-MOVE `identifier` is valid.

    Parameters
    ----------
    identifier : pydicom.dataset.Dataset
        The *Identifier* dataset to check.
    model : pydicom.uid.UID
        The Query/Retrieve Information Model.

    Raises
    ------
    InvalidIdentifier
        If the Identifier is invalid.
    """
    # Part 4, C.4.1.1.3.1, C.4.2.1.4 and C.4.3.1.3.1:
    #   (0008,0052) Query Retrieve Level is required in the Identifier
    if "QueryRetrieveLevel" not in identifier:
        raise InvalidIdentifier(
            "The Identifier contains no Query Retrieve Level element"
        )

    if model in _PATIENT_ROOT:
        attr = _PATIENT_ROOT[model]
    else:
        attr = _STUDY_ROOT[model]

    levels = list(attr.keys())
    if identifier.QueryRetrieveLevel not in levels:
        raise InvalidIdentifier(
            "The Identifier's Query Retrieve Level value is invalid"
        )

    if len(identifier) == 1:
        raise InvalidIdentifier("The Identifier contains no keys")

    for ii, level in enumerate(levels):
        if level == identifier.QueryRetrieveLevel:
            # Check if identifier has elements below current level
            for sublevel in levels[ii + 1 :]:
                if any([kw in identifier for kw in attr[sublevel]]):
                    raise InvalidIdentifier(
                        "The Identifier contains keys below the level "
                        "specified by the Query Retrieve Level"
                    )

            # The level is the same as that in the identifier so we're OK
            return

        # The level is above that in the identifier so make sure the unique
        #   keyword is present
        if attr[level][0] not in identifier:
            raise InvalidIdentifier(
                f"The Identifier is missing a unique key for " f"the '{level}' level"
            )


def clear(session):
    """Delete all entries from the database.

    Parameters
    ----------
    session : sqlalchemy.orm.session.Session
        The session we are using to clear the database.
    """
    for instance in session.query(Instance).all():
        session.delete(instance)

    session.commit()


def create(db_location, echo=False):
    """Create a new database at `db_location` if one doesn't already exist.

    Parameters
    ----------
    db_location : str
        The location of the database.
    echo : bool, optional
        Turn the sqlalchemy logging on (default ``False``).
    """
    engine = create_engine(db_location, echo=echo)

    # Create the tables (won't recreate tables already present)
    Base.metadata.create_all(engine)

    return engine


def remove_instance(instance_uid, session):
    """Remove a SOP Instance from the database.

    Parameters
    ----------
    instance_uid : pydicom.uid.UID
        The (0008,0018) *SOP Instance UID* of the SOP Instance to be removed
        from the database.
    session : sqlalchemy.orm.session.Session
        The session to use when querying the database for the instance.
    """
    matches = (
        session.query(Instance).filter(Instance.sop_instance_uid == instance_uid).all()
    )
    if matches:
        session.delete(matches[0])
        session.commit()


def search(model, identifier, session):
    """Search the database.

    Optional keys are not supported.

    Parameters
    ----------
    model : pydicom.uid.UID
        The Query/Retrieve Information Model. Supported models are:

        - *Patient Root Query Retrieve Information Model* for C-FIND, C-GET
          and C-MOVE
        - *Study Root Query Retrieve Information Model* for C-FIND, C-GET and
          C-MOVE
    identifier : pydicom.dataset.Dataset
        The Query/Retrieve request's *Identifier* dataset.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.

    Returns
    -------
    list of Instance
        The matching database Instances.

    Raises
    ------
    ValueError
        If the `identifier` is invalid.
    """
    if model not in _STUDY_ROOT and model not in _PATIENT_ROOT:
        raise ValueError(f"Unknown information model '{model.name}'")

    # Remove all optional keys, after this only unique/required will remain
    for elem in identifier:
        kw = elem.keyword
        if kw != "QueryRetrieveLevel" and kw not in _ATTRIBUTES:
            delattr(identifier, kw)

    if model in _C_GET or model in _C_MOVE:
        # Part 4, C.2.2.1.2: remove required keys from C-GET/C-MOVE
        for kw, value in _ATTRIBUTES.items():
            if value[1] == "R" and kw in identifier:
                delattr(identifier, kw)

    return _search_qr(model, identifier, session)


def _search_qr(model, identifier, session):
    """Search the database using a Query/Retrieve *Identifier* query.

    Parameters
    ----------
    model : pydicom.uid.UID
        Either *Patient Root Query Retrieve Information Model* or *Study Root
        Query Retrieve Information Model* for C-FIND, C-GET or C-MOVE.
    identifier : pydicom.dataset.Dataset
        The request's *Identifier* dataset.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.

    Returns
    -------
    list of db.Instance
        The Instances that match the query.
    """
    # Will raise InvalidIdentifier if check failed
    _check_identifier(identifier, model)

    if model in _PATIENT_ROOT:
        attr = _PATIENT_ROOT[model]
    else:
        attr = _STUDY_ROOT[model]

    # Hierarchical search method: C.4.1.3.1.1
    query = None
    for level, keywords in attr.items():
        # Keywords at current level that are in the identifier
        keywords = [kw for kw in keywords if kw in identifier]
        # Create query dataset for only the current level and run it
        ds = Dataset()
        [setattr(ds, kw, getattr(identifier, kw)) for kw in keywords]
        query = build_query(ds, session, query)

        if level == identifier.QueryRetrieveLevel:
            break

    return query.all()


def _search_range(elem, session, query=None):
    """Perform a range search for DA, DT and TM elements with '-' in them.

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The attribute to perform the search with.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        An existing query within which this search should be performed. If
        not used then all the Instances in the database will be searched
        (default).

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    # range matching
    #   <date1> - <date2>: matches any date within the range, inclusive
    #   - <date2>: match all dates prior to and including <date2>
    #   <date1> -: match all dates after and including <date1>
    #   <time>: if Timezone Offset From UTC included, values are in specified
    #   date: 20060705-20060707 + time: 1000-1800 matches July 5, 10 am to
    #       July 7, 6 pm.
    start, end = elem.value.split("-")
    attr = getattr(Instance, _TRANSLATION[elem.keyword])
    if not query:
        query = session.query(Instance)

    if start and end:
        return query.filter(attr >= start, attr <= end)
    elif start and not end:
        return query.filter(attr >= start)
    elif not start and end:
        return query.filter(attr <= end)

    raise ValueError("Invalid attribute value for range matching")


def _search_single_value(elem, session, query=None):
    """Perform a search using single value matching.

    Single value matching shall be performed if the value of an Attribute is
    non-zero length and the VR is not SQ and:

    * the VR is AE, CS, LO, LT, PN, SH, ST, UC, UR or UT and contains no wild
      card characters, or
    * the VR is DA, TM or DT and contains a single value with no "-", or
    * any other VR

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The attribute to perform the search with.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        An existing query within which this search should be performed. If
        not used then all the Instances in the database will be searched
        (default).

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    attr = getattr(Instance, _TRANSLATION[elem.keyword])
    if elem.VR == "PN":
        value = str(elem.value)
    else:
        value = elem.value

    if not query:
        query = session.query(Instance)

    return query.filter(attr == value)


def _search_uid_list(elem, session, query=None):
    """Search using an element containing a list of UIDs.

    A match against any of the UIDs is considered a positive result.

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The attribute to perform the search with.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        An existing query within which this search should be performed. If
        not used then all the Instances in the database will be searched
        (default).

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    if not elem.value:
        return _search_universal(elem, session, query)

    attr = getattr(Instance, _TRANSLATION[elem.keyword])
    if not query:
        query = session.query(Instance)

    return query.filter(attr.in_(elem.value))


def _search_universal(elem, session, query=None):
    """Perform a universal search for empty elements.

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The attribute to perform the search with.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        An existing query within which this search should be performed. If
        not used then all the Instances in the database will be searched
        (default).

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    # If the value is zero length then all entities shall match
    if not query:
        query = session.query(Instance)

    return query


def _search_wildcard(elem, session, query=None):
    """Perform a wildcard search.

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The attribute to perform the search with.
    session : sqlalchemy.orm.session.Session
        The session we are using to query the database.
    query : sqlalchemy.orm.query.Query, optional
        An existing query within which this search should be performed. If
        not used then all the Instances in the database will be searched
        (default).

    Returns
    -------
    sqlalchemy.orm.query.Query
        The resulting query.
    """
    # Contains '*' or '?', case-sensitive if not PN
    #   '*' shall match any sequence of characters (incl. zero length)
    #   '?' shall match any single character
    attr = getattr(Instance, _TRANSLATION[elem.keyword])
    if elem.VR == "PN":
        value = str(elem.value)
    else:
        value = elem.value

    if value is None or value == "":
        value = "*"

    value = value.replace("*", "%")
    value = value.replace("?", "_")

    if not query:
        query = session.query(Instance)

    return query.filter(attr.like(value))


# Database table setup stuff
Base = declarative_base()


class Image(Base):
    __tablename__ = "image"
    # (0008,0018) SOP Instance UID | VR UI, VM 1, U
    sop_instance_uid = Column(String(64), primary_key=True)
    # (0020,0013) Instance Number | VR IS, VM 1, R
    instance_number = Column(Integer)


class Instance(Base):
    __tablename__ = "instance"

    # Absolute path to the stored SOP Instance
    filename = Column(String)
    # Transfer Syntax UID of the SOP Instance
    transfer_syntax_uid = Column(String(64))
    sop_class_uid = Column(String(64))

    patient_id = Column(String, ForeignKey("patient.patient_id"))
    patient_name = Column(String, ForeignKey("patient.patient_id"))

    study_instance_uid = Column(String, ForeignKey("study.study_instance_uid"))
    study_date = Column(String, ForeignKey("study.study_date"))
    study_time = Column(String, ForeignKey("study.study_time"))
    accession_number = Column(String, ForeignKey("study.accession_number"))
    study_id = Column(String, ForeignKey("study.study_id"))

    series_instance_uid = Column(String, ForeignKey("series.series_instance_uid"))
    modality = Column(String, ForeignKey("series.modality"))
    series_number = Column(String, ForeignKey("series.series_number"))

    sop_instance_uid = Column(
        String,
        ForeignKey("image.sop_instance_uid"),
        primary_key=True,
    )
    instance_number = Column(String, ForeignKey("image.instance_number"))

    def as_identifier(self, identifier, model):
        """Return an Identifier dataset matching the elements from a query.

        Parameters
        ----------
        identifier : pydicom.dataset.Dataset
            The C-FIND, C-GET or C-MOVE request's *Identifier* dataset.
        model : pydicom.uid.UID
            The Query/Retrieve Information Model.

        Returns
        -------
        pydicom.dataset.Dataset
            The response *Identifier*.
        """
        ds = Dataset()
        ds.QueryRetrieveLevel = identifier.QueryRetrieveLevel

        if model in _PATIENT_ROOT:
            attr = _PATIENT_ROOT[model]
        else:
            attr = _STUDY_ROOT[model]

        all_keywords = []
        for level, keywords in attr.items():
            all_keywords.extend(keywords)
            if level == identifier.QueryRetrieveLevel:
                break

        for kw in [kw for kw in all_keywords if kw in identifier]:
            try:
                attribute = _TRANSLATION[kw]
            except KeyError:
                continue

            setattr(ds, kw, getattr(self, attribute, None))

        return ds

    @property
    def context(self):
        """Return a presentation context for the Instance.

        Returns
        -------
        pynetdicom.presentation.PresentationContext

        Raises
        ------
        ValueError
            If either of the *SOP Class UID* or *Transfer Syntax UID* is not
            available for the Instance.
        """
        if None in [self.sop_class_uid, self.transfer_syntax_uid]:
            raise ValueError(
                "Cannot determine which presentation context is required for "
                "for the SOP Instance"
            )

        return build_context(self.sop_class_uid, self.transfer_syntax_uid)


class Patient(Base):
    __tablename__ = "patient"
    # (0010,0020) Patient ID | VR LO, VM 1, U
    patient_id = Column(String(16), primary_key=True)
    # (0010,0010) Patient's Name | VR PN, VM 1, R
    patient_name = Column(String(64))


class Series(Base):
    __tablename__ = "series"
    # (0020,000E) Series Instance UID | VR UI, VM 1, U
    series_instance_uid = Column(String(64), primary_key=True)
    # (0008,0060) Modality | VR CS, VM 1, R
    modality = Column(String(16))
    # (0020,0011) Series Number | VR IS, VM 1, R
    series_number = Column(Integer)


class Study(Base):
    __tablename__ = "study"
    # (0020,000D) Study Instance UID | VR UI, VM 1, U
    study_instance_uid = Column(String(64), primary_key=True)
    # (0008,0020) Study Date | VR DA, VM 1, R
    study_date = Column(String(8))
    # (0008,0030) Study Time | VR TM, VM 1, R
    study_time = Column(String(14))
    # (0008,0050) Accession Number | VR SH, VM 1, R
    accession_number = Column(String(16))
    # (0020,0010) Study ID | VR SH, VM 1, R
    study_id = Column(String(16))



import socket
from struct import pack
import time


def create_socket(address):
    # AF_INET: IPv4, SOCK_STREAM: TCP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # SO_REUSEADDR: reuse the socket in TIME_WAIT state without
    #   waiting for its natural timeout to expire
    #   Allows local address reuse
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    # If no timeout is set then recv() will block forever if
    #   the connection is kept alive with no data sent
    # SO_RCVTIMEO: the timeout on receive calls in seconds
    #   set using a packed binary string containing two uint32s as
    #   (seconds, microseconds)
    # if self.assoc.network_timeout is not None:
    timeout_seconds = 5
    timeout_microsec = 5 % 1 * 1000
    sock.setsockopt(
        socket.SOL_SOCKET,
        socket.SO_RCVTIMEO,
        pack("ll", timeout_seconds, timeout_microsec),
    )

    sock.bind(address)

    return sock


if __name__ == "__main__":
    for ii in range(60):
        sock = create_socket(("localhost", 0))
        sock.connect(("localhost", 11112))
        sock.settimeout(None)

        try:
            sock.shutdown(socket.SHUT_RDWR)
        except:
            pass
        sock.close()

        time.sleep(10)



"""Implements the supported Service Classes that make use of DIMSE-N."""

import logging
from typing import TYPE_CHECKING, Union

from pynetdicom.dimse_primitives import (
    N_ACTION,
    N_CREATE,
    N_DELETE,
    N_EVENT_REPORT,
    N_GET,
    N_SET,
    C_FIND,
)
from pynetdicom.service_class import ServiceClass
from pynetdicom.status import (
    GENERAL_STATUS,
    APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS,
    MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS,
    PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS,
    PROCEDURE_STEP_STATUS,
    STORAGE_COMMITMENT_SERVICE_CLASS_STATUS,
    RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS,
    UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS,
)

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.presentation import PresentationContext

    _MCM = Union[N_CREATE, N_GET, N_ACTION]
    _PJ = Union[N_CREATE, N_EVENT_REPORT, N_GET, N_SET, N_ACTION, N_DELETE]
    _PS = Union[N_CREATE, N_EVENT_REPORT, N_GET, N_SET]
    _SCS = Union[N_EVENT_REPORT, N_ACTION]
    _UPS = Union[N_CREATE, N_EVENT_REPORT, N_GET, N_SET, N_ACTION, C_FIND]


LOGGER = logging.getLogger("pynetdicom.service-n")


class ApplicationEventLoggingServiceClass(ServiceClass):
    """Implementation of the Application Event Logging Service Class

    .. versionadded:: 1.4
    """

    statuses = APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS

    def SCP(self, req: N_ACTION, context: "PresentationContext") -> None:
        """The SCP implementation for Application Event Logging Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_ACTION
            The N-ACTION request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Application Event Logging"
            )


class DisplaySystemManagementServiceClass(ServiceClass):
    """Implementation of the Display System Management Service Class."""

    statuses = GENERAL_STATUS

    def SCP(self, req: N_GET, context: "PresentationContext") -> None:
        """The SCP implementation for Display System Management.

        Parameters
        ----------
        req : dimse_primitives.N_GET
            The N-GET request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_GET):
            self._n_get_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Display System Management"
            )


class InstanceAvailabilityNotificationServiceClass(ServiceClass):
    """Implementation of the Instance Availability Service Class

    .. versionadded:: 1.4
    """

    statuses = GENERAL_STATUS

    def SCP(self, req: N_CREATE, context: "PresentationContext") -> None:
        """The SCP implementation for Instance Availability Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE
            The N-CREATE request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_CREATE):
            self._n_create_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Instance Availability"
            )


class MediaCreationManagementServiceClass(ServiceClass):
    """Implementation of the Media Creation Management Service Class

    .. versionadded:: 1.4
    """

    statuses = MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS

    def SCP(self, req: "_MCM", context: "PresentationContext") -> None:
        """The SCP implementation for Media Creation Management Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE or N_GET or N_ACTION
            The N-CREATE, N-GET or N-ACTION request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_CREATE):
            self._n_create_scp(req, context)
        elif isinstance(req, N_GET):
            self._n_get_scp(req, context)
        elif isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Media Creation Management"
            )


class PrintManagementServiceClass(ServiceClass):
    """Implementation of the Print Management Service Class

    .. versionadded:: 1.4
    """

    statuses = PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS

    def SCP(self, req: "_PJ", context: "PresentationContext") -> None:
        """The SCP implementation for Print Management Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE or N_SET or N_DELETE or N_GET or N_EVENT_REPORT or N_ACTION
            The N-CREATE, N-SET, N-GET, N-DELETE, N-ACTION or N-EVENT-REPORT
            request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_CREATE):
            self._n_create_scp(req, context)
        elif isinstance(req, N_EVENT_REPORT):
            self._n_event_report_scp(req, context)
        elif isinstance(req, N_GET):
            self._n_get_scp(req, context)
        elif isinstance(req, N_SET):
            self._n_set_scp(req, context)
        elif isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        elif isinstance(req, N_DELETE):
            self._n_delete_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Print Management"
            )


class ProcedureStepServiceClass(ServiceClass):
    """Implementation of the Modality Performed Procedure Step Service Class

    .. versionadded:: 1.3
    """

    statuses = PROCEDURE_STEP_STATUS

    def SCP(self, req: "_PS", context: "PresentationContext") -> None:
        """The SCP implementation for Modality Performed Procedure Step.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE or N_SET or N_GET or N_EVENT_REPORT
            The N-CREATE, N-SET, N-GET or N-EVENT-REPORT request primitive
            sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_CREATE):
            # Modality Performed Procedure Step
            self._n_create_scp(req, context)
        elif isinstance(req, N_EVENT_REPORT):
            # Modality Performed Procedure Step Notification
            self._n_event_report_scp(req, context)
        elif isinstance(req, N_GET):
            # Modality Performed Procedure Step Retrieve
            self._n_get_scp(req, context)
        elif isinstance(req, N_SET):
            # Modality Performed Procedure Step
            self._n_set_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Modality Performed Procedure Step"
            )


class RTMachineVerificationServiceClass(ServiceClass):
    """Implementation of the RT Machine Verification Service Class

    .. versionadded:: 1.4
    """

    statuses = RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS

    def SCP(self, req: "_PJ", context: "PresentationContext") -> None:
        """The SCP implementation for RT Machine Verification Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE or N_SET or N_DELETE or N_GET or N_EVENT_REPORT or N_ACTION
            The N-CREATE, N-SET, N-GET, N-DELETE, N-ACTION or N-EVENT-REPORT
            request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_CREATE):
            self._n_create_scp(req, context)
        elif isinstance(req, N_EVENT_REPORT):
            self._n_event_report_scp(req, context)
        elif isinstance(req, N_GET):
            self._n_get_scp(req, context)
        elif isinstance(req, N_SET):
            self._n_set_scp(req, context)
        elif isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        elif isinstance(req, N_DELETE):
            self._n_delete_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with RT Machine Verification"
            )


class StorageCommitmentServiceClass(ServiceClass):
    """Implementation of the Storage Commitment Service Class

    .. versionadded:: 1.4
    """

    statuses = STORAGE_COMMITMENT_SERVICE_CLASS_STATUS

    def SCP(self, req: "_SCS", context: "PresentationContext") -> None:
        """The SCP implementation for Storage Commitment Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_EVENT_REPORT or N_ACTION
            The N-ACTION or N-EVENT-REPORT request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        if isinstance(req, N_EVENT_REPORT):
            self._n_event_report_scp(req, context)
        elif isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Storage Commitment"
            )


class UnifiedProcedureStepServiceClass(ServiceClass):
    """Implementation of the Unified Procedure Step Service Class

    .. versionadded:: 1.4
    """

    statuses = UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS

    def SCP(self, req: "_UPS", context: "PresentationContext") -> None:
        """The SCP implementation for Unified Procedure Step Service Class.

        Parameters
        ----------
        req : dimse_primitives.N_CREATE or C_FIND or N_SET or N_GET or N_EVENT_REPORT or N_ACTION
            The N-CREATE, C-FIND, N-SET, N-GET, N-ACTION or N-EVENT-REPORT
            request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        """
        # UPS Push: N-CREATE, N-ACTION, N-GET
        # UPS Pull: C-FIND, N-GET, N-SET, N-ACTION
        # UPS Watch: N-ACTION, N-GET, C-FIND
        # UPS Event: N-EVENT-REPORT
        if isinstance(req, N_CREATE):
            self._n_create_scp(req, context)
        elif isinstance(req, N_EVENT_REPORT):
            self._n_event_report_scp(req, context)
        elif isinstance(req, N_GET):
            self._n_get_scp(req, context)
        elif isinstance(req, N_SET):
            self._n_set_scp(req, context)
        elif isinstance(req, N_ACTION):
            self._n_action_scp(req, context)
        elif isinstance(req, C_FIND):
            self._c_find_scp(req, context)
        else:
            raise ValueError(
                f"Invalid DIMSE primitive '{req.__class__.__name__}' used "
                f"with Unified Procedure Step"
            )
